# -*- coding: utf-8 -*-
"""
Created on Sat Jan  1 22:33:14 2022

@author: msahr
"""

# -*- coding: utf-8 -*-
"""
Created on Mon Dec 20 14:14:37 2021

@author: msahr
"""

import extractDef as ext
import matplotlib.pyplot as plt
import sys
import re
import math

class RuleList:
#
# DESCRIPTION: A rule list is a linked list of RuleAtom elements. Each RuleAtom 
# describes one drc rule.
#
    def __init__(self):
        self.Atom=None
        self.next=None
    def insert(self,rule):
#
# DESCRIPTION: This function inserts rule at the end of the linked list headed by
# self.Atom. If self.Atom==None it points to self.Atom-rule
#
# INPUT: RuleAtom Element
#
# OUTPUT:None
#
# KNOWN ISSUES: NONE
#
        if self.Atom != None:
            newrule=RuleList()
            newrule.Atom=rule
            ruleActive=self
            while ruleActive.next is not None:
                ruleActive=ruleActive.next
            ruleActive.next=newrule
        else:
            self.Atom=rule

        
class RuleAtom:
#
# DESCRIPTION: RuleAtom is the basic drc rule element. It consists of the name of the rule,
# layers: layer1, layer2 which are ascii text labels with layer names for example "M0", "M1" etc,
# Two floating values, val1, val2 that contains the limit of the rule and two relations that can
# be either '=','<','>','<=','>='. 
#
    def __init__(self):
        self.name=''
        self.layer1=''
        self.layer2=''
        self.val1=-1
        self.val2=-1
        self.relation1=''
        self.relation2=''
        
def read_drcrules(rules):
#
# INPUT: text file containif the DRC rules
#
# OUTPUT:A linked list of the rules
#
# KNOWN ISSUES: NONE
#
    try:
        myfile=open(rules,'r')
    except:
        print('DRC rule file',rules,' not found')
        sys.exit()
    RuleCount=0
    RuleON=False
    line=myfile.readline()
    RuleHead=RuleList()
    while line !='END\n' :
        if not RuleON:
            if re.split(' ',line)[0]=='Rule':
                RuleCount+=1
                Rule=RuleAtom()
                Name=re.split('[ \n]',line)[1]
                Rule.name=Name
                RuleON=True
                RuleHead.insert(Rule)
        else:
            if re.split('\t| |,',line)[1]=='Width':
                Type='Width'
                NumberOfLayers=1
            elif re.split('\t| |,',line)[1]=='Space':
                Type='Space'
                NumberOfLayers=1
            elif re.split('\t| |,',line)[1]=='Enclosure':
                Type='Enclosure'
                NumberOfLayers=2
            elif re.split('\t| |,',line)[1]=='Overlap':
                Type='Overlap'
                NumberOfLayers=2
            elif re.split('\t| |,',line)[1]=='Antenna':
                Type='Antenna'
                NumberOfLayers=1
            elif re.split('\t| |,',line)[1]=='Density':
                Type='Density'
                NumberOfLayers=1
            elif re.split('\t| |,',line)[1]=='Notch':
                Type='Notch'
                NumberOfLayers=1
            else:
                print('DRC type not recognized:',re.split('\t| |,',line)[1])
                sys.exit(0)
            Layer1=re.split('\t| |,',line)[2]
            Rule.layer1=Layer1
            if NumberOfLayers==2:
                Layer2=re.split('\t| |,',line)[3]
                Rule.layer2=Layer2
            else:
                Layer2=''
            Rule.Type=Type
            ReadingRule=True
            if NumberOfLayers==2:
                item=4
            else:
                item=3
            if re.split('\t| |,',line)[item][0]=='[':
                Condition=re.split('\t|,|\[|\]',line)[2]
                if re.split(' ',Condition)[0]=='Width':
                    if Type=='Space':
                        Rule.Type=Type='Space_Width'
                    else:
                        print('Syntax Error in DRC rule:',Rule.name,line)
                        print('The conditional Width can only be used together with Space rules')
                        sys.exit(0)
                elif re.split(' ',Condition)[0]=='Runlength':
                    if Type=='Space':
                        Rule.Type=Type='Space_Runlength'
                    else:
                        print('Syntax Error in DRC rule:',Rule.name,line)
                        print('The conditional Runlength can only be used together with Space rules')
                        sys.exit(0)
                if re.split(' ',Condition)[1] != '>':
                # if re.split(' ',Condition)[1]=='>':
                #     Rule.relation2='>'
                # elif re.split(' ',Condition)[1]=='>=':
                #     Rule.relation2='>='
                # elif re.split(' ',Condition)[1]=='<':
                #     Rule.relation2='<'
                # elif re.split(' ',Condition)[1]=='<=':
                #     Rule.relation2='<='
                # elif re.split(' ',Condition)[1]=='=':
                #     Rule.relation2='='
                # else:
                    print('Syntax Error in DRC rule:',Rule.name,line)
                    print('Only > allowed in Condition statment',re.split('\[|\]| ',Condition)[2])
                    sys.exit(0)
                Rule.val2=float(re.split(' ',Condition)[2])
                item+=3
            while ReadingRule:
                if re.split('\t| |,',line)[item]=='>':
                    Rule.relation1='>'
                    item=item+1
                elif re.split('\t| |,',line)[item]=='>=':
                    Rule.relation1='>='
                    item=item+1
                elif re.split('\t| |,',line)[item]=='<':
                    Rule.relation1='<'
                    item=item+1
                elif re.split('\t| |,',line)[item]=='<=':
                    Rule.relation1='<='
                    item=item+1
                elif re.split('\t| |,',line)[item]=='=':
                    Rule.relation1='='
                    item=item+1
                else:
                    Rule.relation1='='
                Rule.val1=float(re.split('\t| |,',line)[item])
                item=item+1
                try:
                    relation=re.split('\t| |,',line)[item]
                except IndexError:
                    ReadingRule=False
                    RuleON=False
                if RuleON:
                    if relation != '&':
                        RuleCount+=1
                        Rule=RuleAtom()
                        Rule.name=Name
                        Rule.layer1=Layer1
                        Rule.layer2=Layer2
                        Rule.Type=Type
                        RuleHead.insert(Rule)
                    else:
    #
    # A '&' means that two relations need to be combined as in width>10 & width < 20. 
    # The '&' makes sure both relationships needs to pass in order to satisfy the rule
    # This combination of two relationships is embodied in the Rule.relation1, 2 and Rule.val1,val2
    # in the RuleAtom definition
    #
                        Rule.relation2=Rule.relation1
                        Rule.val2=Rule.val1
                        item=item+1
                        try:
                            re.split('\t| |,',line)[item]
                        except IndexError:
                            print('Syntax Error in rule:',line, 'No character after &')
                            sys.exit(1)
        line=myfile.readline()                  
    return RuleHead

def verify_drc_rule(value, rule):
# 
# DESCRIPTION: The rule defined by input rule is verified with the input value
# INPUT: A floating point; Value, A RuleAtom=rule
#
# OUTPUT: True/False
#
# KNOWN ISSUES: NONE
#
    ruleActive=rule
    ActiveRuleName=rule.Atom.name
    RuleMet=False
    while ruleActive is not None and ruleActive.Atom.name==ActiveRuleName and not RuleMet:
        if ruleActive.Atom.relation1=='=':
            if value==ruleActive.Atom.val1:
                RuleMet=True
        elif ruleActive.Atom.relation1=='>':
            if value>ruleActive.Atom.val1:
                RuleMet=True
        elif ruleActive.Atom.relation1=='>=':
            if value>=ruleActive.Atom.val1:
                RuleMet=True
        elif ruleActive.Atom.relation1=='<':
            if value<ruleActive.Atom.val1:
                RuleMet=True
        elif ruleActive.Atom.relation1=='<=':
            if value<=ruleActive.Atom.val1:
                RuleMet=True
        else:
            print('Unknown rule relation',ruleActive.Atom.relation)
            sys.exit(0)
        if RuleMet and ruleActive.Atom.relation2 != '':
            RuleMet=False
            if ruleActive.Atom.relation2=='=':
                if value==ruleActive.Atom.val2:
                    RuleMet=True
            elif ruleActive.Atom.relation2=='>':
                if value>ruleActive.Atom.val2:
                    RuleMet=True
            elif ruleActive.Atom.relation2=='>=':
                if value>=ruleActive.Atom.val2:
                    RuleMet=True
            elif ruleActive.Atom.relation2=='<':
                if value<ruleActive.Atom.val2:
                    RuleMet=True
            elif ruleActive.Atom.relation2=='<=':
                if value<=ruleActive.Atom.val2:
                    RuleMet=True
            else:
                print('Unknown rule relation',ruleActive.Atom.relation)
                sys.exit(0)
        ruleActive=ruleActive.next
    return RuleMet



def polygon_width(p):
#
# DESCRIPTION: The width of a polygon is calculated as the shortest distance
# across it.
# INPUT: A polygon: p
#
# OUTPUT: Width of the polygon as a floating point number
#
# KNOWN ISSUES: NONE
#
    width=ext.MAXV
    pActive=p.head
    midpnt=ext.Point()
    line=ext.Line()
    while pActive.next is not p.head:
        midpnt.x=0.5*(pActive.pnt.x+pActive.next.pnt.x)
        midpnt.y=0.5*(pActive.pnt.y+pActive.next.pnt.y)
        line.pnt1.x=midpnt.x
        line.pnt1.y=midpnt.y
        line.pnt2.x=pActive.next.pnt.x
        line.pnt2.y=pActive.next.pnt.y
        lrot=ext.rotate_line(line,90)
# Resize line to ensure it covers polygon
        Resize=(p.BB.urx-p.BB.llx+p.BB.ury-p.BB.lly)/math.sqrt((lrot.pnt2.x)**2+(lrot.pnt2.y)**2)
        lrot.pnt2.x=lrot.pnt2.x*Resize
        lrot.pnt2.y=lrot.pnt2.y*Resize
# rotate return a line starting at (0,0)
        lrot.pnt1.x=midpnt.x
        lrot.pnt1.y=midpnt.y
        lrot.pnt2.x=lrot.pnt2.x+midpnt.x
        lrot.pnt2.y=lrot.pnt2.y+midpnt.y
        pnt=ext.find_closest_line_polygon_intersect_pnt(lrot,p)
        if pnt is not None:
            span=math.sqrt( (lrot.pnt1.x-pnt.x)**2+(lrot.pnt1.y-pnt.y)**2 )
            if span < width:
                MinSpanPnt=ext.Point()
                MinSpanPnt.x=pnt.x
                MinSpanPnt.y=pnt.y
                MinLine=ext.Line()
                MinLine.pnt1.x=pActive.pnt.x
                MinLine.pnt1.y=pActive.pnt.y
                MinLine.pnt2.x=pActive.next.pnt.x
                MinLine.pnt2.y=pActive.next.pnt.y
                width=span
        pActive=pActive.next
    return width

def drc_space_width(p1, p2, DrcRuleList, Layer1, Layer2):
    # 
# DESCRIPTION: A spacing rule based on polygon width is implemented
#   The routine checks both polygons to see if the have the required width
#   and, if so the associated spacing rule is checked
#
# INPUT: Two polygons, p1, p2; The rule: DrcRuleList. The
#   layers belonging to the two polygons: Layer1, Layer2
#
# OUTPUT:A linked list of Error Polygons
#
# KNOWN ISSUES: Only the Layer belonging to the first polygon is verified
#
    ErrorPolygonsList=None
    ruleActive=DrcRuleList
    while ruleActive is not None:
        if ruleActive.Atom.Type=='Space_Width' and ruleActive.Atom.layer1 == Layer1:# Need to include and ruleActive.Atom.layer2 == Layer2:
            if polygon_width(p1) > ruleActive.Atom.val2 or polygon_width(p2) > ruleActive.Atom.val2:
                ErrorPolygonsList=drc_space(p1,p2,DrcRuleList,Layer1,Layer2)
        ruleActive=ruleActive.next                         
    if ErrorPolygonsList is not None:
        return ErrorPolygonsList
    else:
        return None 
    
def drc_antenna(net, DrcRuleList):
    # net is a list of polygonlist
#
# INPUT: net is a list of polygonlist, The rule: DrcRuleList 
#
# OUTPUT:A linked list of ErrorPolygons
#
# KNOWN ISSUES: Only circumference and area calculated. No metal jumo checks
# are implemented
#
    ErrorPolygonsList=ext.PolygonList()  
    netListActive=net
    FoundFET=False
    while netListActive is not None and not FoundFET:
        if netListActive.Name == 'PMOSbody' or netListActive.Name == 'NMOSbody':
            FoundFET=True
        netListActive=netListActive.next
    if not FoundFET:
        return None
    ruleActive=DrcRuleList
    while ruleActive is not None:
        if ruleActive.Atom.Type=='Antenna':
            netListActive=net
            FoundTopMetal=False
            while netListActive is not None and not FoundTopMetal:
                if netListActive.Name == ruleActive.Atom.layer1:
                    FoundTopMetal=True
                else:
                    netListActive=netListActive.next
            if not FoundTopMetal:
                return None
            MetalActive=netListActive.PolyList
            circumference=0
            BB=ext.BoundingBox()
            BB.llx=BB.lly=ext.MAXV
            BB.urx=BB.ury=ext.MINV
            while MetalActive is not None:
                BB.llx=ext.MIN(BB.llx,MetalActive.Polygon.BB.llx)
                BB.lly=ext.MIN(BB.lly,MetalActive.Polygon.BB.lly)
                BB.urx=ext.MAX(BB.urx,MetalActive.Polygon.BB.urx)
                BB.ury=ext.MAX(BB.ury,MetalActive.Polygon.BB.ury)
                circumference+=ext.polygon_circumference(MetalActive.Polygon)
                MetalActive=MetalActive.next
            FoundDiffusion=False
            netListActive=net
            while netListActive is not None and not FoundDiffusion:
                if netListActive.Name == 'Diffusion':
                    FoundDiffusion=True
                else:
                    netListActive=netListActive.next
            DiffusionActive=netListActive.PolyList
            Area=0
            while DiffusionActive is not None:
                Area+=ext.polygon_area(DiffusionActive.Polygon)
                DiffusionActive=DiffusionActive.next
            AntennaRatio=circumference/Area               
            if verify_drc_rule(AntennaRatio,ruleActive) == False:
                pnt1=ext.Point()
                pnt2=ext.Point()
                pnt3=ext.Point()
                pnt1.x=BB.llx
                pnt1.y=BB.lly
                pnt2.x=BB.urx
                pnt2.y=BB.lly
                pnt3.x=BB.urx
                pnt3.y=BB.ury
                ErrorPolygonsList.insert(ext.create_3pnt_polygon(pnt1,pnt2,pnt3))
        ruleActive=ruleActive.next    
    if ErrorPolygonsList.Polygon is not None:
        return ErrorPolygonsList
    else:
        return None 

def drc_space_runlength(p1, p2, DrcRuleList, Layer1, Layer2):
#
# DESCRIPTION: A spacing rule based on polygon width is implemented
#   The routine checks both polygons to see if the have the required width
#   and, if so the associated spacing rule is checked
#
# INPUT: Two polygons, p1, p2; The rule: DrcRuleList. The
#   layers belonging to the two polygons: Layer1, Layer2
#
# OUTPUT:A linked list of Error Polygons
#
# KNOWN ISSUES: Only the Layer belonging to the first polygon is verified
#
    ErrorPolygonsList=ext.PolygonList()    
    p1Active=p1.head
    First1=True
    while p1Active is not p1.head or First1:
        First1=False
        p2Active=p2.head
        First2=True
        while p2Active != p2.head or First2:
            First2=False
            runl,space=runlength(p1Active,p2Active)
            if runl < ext.MAXV:
                ruleActive=DrcRuleList
                while ruleActive is not None:
                    if ruleActive.Atom.Type=='Space_Runlength' and ruleActive.Atom.layer1 == Layer1:# and ruleActive.Atom.layer2 == Layer2:
                        if runl> ruleActive.Atom.val2 :
                            if verify_drc_rule(space,ruleActive) == False:
                                ErrorPolygonsList.insert(ext.create_3pnt_polygon(p1Active.pnt,p1Active.next.pnt,p2Active.pnt))
                    ruleActive=ruleActive.next                     
            p2Active=p2Active.next
        p1Active=p1Active.next
    if ErrorPolygonsList.Polygon is not None:
        return ErrorPolygonsList
    else:
        return None 

def drc_density(polyList,DrcRuleList,Layer, BB):
#
# DESCRIPTION: THe density is verified by stepping a Square 600x600 large
# across the area defined by BB. The Square is stepped in steps of half the
# size of the square.
#
# INPUT: A list of polygons:polyList; A rule:DrcRuleList; A polygon layer:Layer1
# that is associated with the polyList and a bounding box:BB describing the required
# area to be covers
#
# OUTPUT:A linked list of Error Polygons
#
# KNOWN ISSUES: NONE
#
    ErrorPolygonsList=ext.PolygonList()
    if DrcRuleList.Atom.Type != 'Density':
        return None
    ErrorPolygonsList=ext.PolygonList()
    SquareSize=600
    NStepsInX=int((BB.urx-BB.llx)/SquareSize*2)
    NStepsInY=int((BB.ury-BB.lly)/SquareSize*2)
    StepSizeInX=SquareSize/2
    StepSizeInY=SquareSize/2
    for i in range(NStepsInX):
        print('i',i,NStepsInX)
        for j in range(NStepsInY):
            Square=ext.Polygon()
            llx=BB.llx+i*StepSizeInX-3
            lly=BB.lly+j*StepSizeInY-3
            urx=llx+SquareSize
            ury=lly+SquareSize
            v1=ext.Vertex(llx,lly)
            v2=ext.Vertex(urx,lly)
            v3=ext.Vertex(urx,ury)
            v4=ext.Vertex(llx,ury)
            Square.insert(Square.head,v1)
            Square.insertLast(v2)
            Square.insertLast(v3)
            Square.insertLast(v4)
            Square.closeLoop()
            # Square.plot('#222222')
            polyActive=polyList
            Count=0
            PolygonsInsideSquare=ext.PolygonList()
            while polyActive is not None:
                polygon=polyActive.Polygon.copy()
                polyAND=ext.poly_AND(polygon,Square)
                if polyAND is not None:
                    Count+=1
                    # polyAND.Polygon.plot('#222222')
                    PolygonsInsideSquare.insert(polyAND.Polygon) # assuming it is only one polygon returned
                # else:
                #     polygon.plot('#222222')
                polyActive=polyActive.next
            Area=0
            polyActive=PolygonsInsideSquare
            while polyActive is not None:
                Area=Area+ext.polygon_area(polyActive.Polygon)
                polyActive=polyActive.next
            Density=Area/SquareSize/SquareSize
            ruleActive=DrcRuleList
            while ruleActive is not None:
                if ruleActive.Atom.Type=='Density' and ruleActive.Atom.layer1 == Layer:
                    if verify_drc_rule(Density,ruleActive) == False:
                        # print('Density',Density)
                        ErrorPolygonsList.insert(Square)
                    RuleName=ruleActive.Atom.name
                    while ruleActive is not None and ruleActive.Atom.name == RuleName:
                        ruleActive=ruleActive.next
                else:
                    ruleActive=ruleActive.next 
    if ErrorPolygonsList.Polygon is not None:
        return ErrorPolygonsList
    else:
        return None 
    

def drc_area(p,DrcRuleList,Layer):
#
# DESCRIPTION: The area of a polygon is verified agains a design rule
#
# INPUT: A polygon:p; Rule:DrcRUleList; The Layer:Layer associated with the polygon p
#
# OUTPUT:A linked list of Error Polygons
#
# KNOWN ISSUES: NONE
#
    ErrorPolygonsList=ext.PolygonList()
    area=ext.polygon_area(p)
    ruleActive=DrcRuleList
    while ruleActive is not None:
        if ruleActive.Atom.Type=='Area' and ruleActive.Atom.layer1 == Layer:
            if verify_drc_rule(area,ruleActive) == False:
                ErrorPolygonsList.insert(ext.create_3pnt_polygon(p.head,p.head.next,p.head.next.next))
            RuleName=ruleActive.Atom.name
            while ruleActive is not None and ruleActive.Atom.name == RuleName:
                ruleActive=ruleActive.next
        else:
            ruleActive=ruleActive.next    
    if ErrorPolygonsList.Polygon is not None:
        return ErrorPolygonsList
    else:
        return None 

def drc_overlap(p1, p2, DrcRuleList, Layer1, Layer2):
#
# DESCRIPTION: An overlap check is impplemented by spot checking the distance from
# the covered polygon to the edge of the covering polygon. The distance is checked halfway 
# along an edge.
#
# INPUT: Two polygons: p1, p2; A rule:DrcRuleList; Two layers: Layer1, Layer2 associated
# with the two polygons.
#
# OUTPUT:A linked list of error polygons
#
# KNOWN ISSUES: The algorithm only spot checks half way across edges.
#
    #
    ErrorPolygonsList=ext.PolygonList()
    p1Active=p1.head
    line=ext.Line()
    midpnt=ext.Point()
    FoundExternalEdge=False
    First1=True
    while p1Active is not p1.head or First1:
        First1=False
        if not ext.point_inside_polygon(p1Active.pnt, p2) and not ext.point_inside_polygon(p1Active.next.pnt, p2):
            FoundExternalEdge=True
            midpnt.x=0.5*(p1Active.pnt.x+p1Active.next.pnt.x)
            midpnt.y=0.5*(p1Active.pnt.y+p1Active.next.pnt.y)
            line.pnt1.x=midpnt.x
            line.pnt1.y=midpnt.y
            line.pnt2.x=p1Active.next.pnt.x
            line.pnt2.y=p1Active.next.pnt.y
            lrot=ext.rotate_line(line,90)
    # Resize line to ensure it covers polygon 2
            Resize=(p2.BB.urx-p2.BB.llx+p2.BB.ury-p2.BB.lly)/math.sqrt((lrot.pnt2.x)**2+(lrot.pnt2.y)**2)
            lrot.pnt2.x=lrot.pnt2.x*Resize
            lrot.pnt2.y=lrot.pnt2.y*Resize
    # rotate return a line starting at (0,0)
            lrot.pnt1.x=midpnt.x
            lrot.pnt1.y=midpnt.y
            lrot.pnt2.x=lrot.pnt2.x+midpnt.x
            lrot.pnt2.y=lrot.pnt2.y+midpnt.y
            pnt=ext.find_closest_line_polygon_intersect_pnt(lrot,p2)
            overlap=math.sqrt( (lrot.pnt1.x-pnt.x)**2+(lrot.pnt1.y-pnt.y)**2 )
            ruleActive=DrcRuleList
            while ruleActive is not None:
                if ruleActive.Atom.Type=='Overlap' and ruleActive.Atom.layer1 == Layer1 and ruleActive.Atom.layer2 == Layer2:
                    if verify_drc_rule(overlap,ruleActive) == False:
                        ErrorPolygonsList.insert(ext.create_3pnt_polygon(line.pnt1,line.pnt2,pnt))
                    RuleName=ruleActive.Atom.name
                    while ruleActive is not None and ruleActive.Atom.name == RuleName:
                        ruleActive=ruleActive.next
                else:
                    ruleActive=ruleActive.next
        p1Active=p1Active.next
    if not FoundExternalEdge:
        ErrorPolygonsList.insert(p1.copy)
    if ErrorPolygonsList.Polygon is not None:
        return ErrorPolygonsList
    else:
        return None       

def drc_enclosure(p1, p2, DrcRuleList, Layer1, Layer2):
#
# DESCRIPTION: An enclosre check is impplemented by spot checking the distance from
# the covered polygon to the edge of the covering polygon. The distance is checked halfway 
# along an edge.
#
# INPUT: Two polygons: p1, p2; A rule:DrcRuleList; Two layers: Layer1, Layer2 associated
# with the two polygons.
#
# OUTPUT:A linked list of error polygons
#
# KNOWN ISSUES: The algorithm only spot checks half way across edges.
#
    ErrorPolygonsList=ext.PolygonList()
    # first let us make sure all vertices in p1 resides inside p2
    if not ext.point_inside_polygon(p1.head.pnt, p2):
        ErrorPolygonsList.insert(ext.create_3pnt_polygon(p1.head.prev,p1.head,p1.head.next))
    p1Active=p1.head.next
    while p1Active is not p1.head:
        if not ext.point_inside_polygon(p1Active.pnt, p2):
            ErrorPolygonsList.insert(ext.create_3pnt_polygon(p1Active.prev.pnt,p1Active.pnt,p1Active.next.pnt))
        p1Active=p1Active.next
    # It might be best to return here if ErrorPolygonsList is not empty !!!
    if ErrorPolygonsList.Polygon is not None:
        return ErrorPolygonsList
    p1Active=p1.head
    line=ext.Line()
    midpnt=ext.Point()
    First1=True
    while p1Active is not p1.head or First1:
        First1=False
        midpnt.x=0.5*(p1Active.pnt.x+p1Active.next.pnt.x)
        midpnt.y=0.5*(p1Active.pnt.y+p1Active.next.pnt.y)
        line.pnt1.x=midpnt.x
        line.pnt1.y=midpnt.y
        line.pnt2.x=p1Active.next.pnt.x
        line.pnt2.y=p1Active.next.pnt.y
        lrot=ext.rotate_line(line,-90)
# Resize line to ensure it covers polygon 2
        Resize=(p2.BB.urx-p2.BB.llx+p2.BB.ury-p2.BB.lly)/math.sqrt((lrot.pnt2.x)**2+(lrot.pnt2.y)**2)
        lrot.pnt2.x=lrot.pnt2.x*Resize
        lrot.pnt2.y=lrot.pnt2.y*Resize
# rotate return a line starting at (0,0)
        lrot.pnt1.x=midpnt.x
        lrot.pnt1.y=midpnt.y
        lrot.pnt2.x=lrot.pnt2.x+midpnt.x
        lrot.pnt2.y=lrot.pnt2.y+midpnt.y
        pnt=ext.find_closest_line_polygon_intersect_pnt(lrot,p2)
        enclosure=math.sqrt( (lrot.pnt1.x-pnt.x)**2+(lrot.pnt1.y-pnt.y)**2 )
        ruleActive=DrcRuleList
        while ruleActive is not None:
            if ruleActive.Atom.Type=='Enclosure' and ruleActive.Atom.layer1 == Layer1 and ruleActive.Atom.layer2 == Layer2:
                if verify_drc_rule(enclosure,ruleActive) == False:
                    ErrorPolygonsList.insert(ext.create_3pnt_polygon(line.pnt1,line.pnt2,pnt))
                RuleName=ruleActive.Atom.name
                while ruleActive is not None and ruleActive.Atom.name == RuleName:
                    ruleActive=ruleActive.next
            else:
                ruleActive=ruleActive.next        
        p1Active=p1Active.next
    if ErrorPolygonsList.Polygon is not None:
        return ErrorPolygonsList
    else:
        return None        


def drc_space(p1, p2, DrcRuleList, Layer1, Layer2):
# 
# DESCRIPTION: Here the spacing between two polygons are verified against a rule
#
# INPUT: Two polygons: p1, p2; A rule:DrcRuleList; Two layers: Layer1, Layer2 associated
# with the two polygons.
#
# OUTPUT:A linked list of error polygons
#
# KNOWN ISSUES: NONE
#
    ErrorPolygonsList=ext.PolygonList()
    p1Active=p1.head
    line=ext.Line()
    First1=True
    while p1Active is not p1.head or First1:
        First1=False
        line.pnt1.x=p1Active.pnt.x
        line.pnt1.y=p1Active.pnt.y
        line.pnt2.x=p1Active.next.pnt.x
        line.pnt2.y=p1Active.next.pnt.y
        p2Active=p2.head
        First2=True
        while p2Active != p2.head or First2:
            First2=False
            v1=ext.Point()
            v1.x=line.pnt2.x-line.pnt1.x
            v1.y=line.pnt2.y-line.pnt1.y
            v2=ext.Point()
            v2.x=p2Active.pnt.x-line.pnt1.x
            v2.y=p2Active.pnt.y-line.pnt1.y
            v3=ext.Point()
            v3.x=p2Active.pnt.x-line.pnt2.x
            v3.y=p2Active.pnt.y-line.pnt2.y
            Acute1=ext.scalar(v1, v2) >= 0
            v1.x=-v1.x
            v1.y=-v1.y
            Acute2=ext.scalar(v1,v3) >=0
            Acute=Acute1 and Acute2
            LineLength=(line.pnt2.x-line.pnt1.x)**2+(line.pnt2.y-line.pnt1.y)**2
            LineP1PntLength=(line.pnt1.x-p2Active.pnt.x)**2+(line.pnt1.y-p2Active.pnt.y)**2
            LineP2PntLength=(line.pnt2.x-p2Active.pnt.x)**2+(line.pnt2.y-p2Active.pnt.y)**2
            if Acute:
                SpacePolygon=ext.create_3pnt_polygon(line.pnt1,line.pnt2,p2Active.pnt)
                Area=abs(ext.polygon_area(SpacePolygon))
                space=Area*2/math.sqrt(LineLength)
                ruleActive=DrcRuleList
                while ruleActive is not None:
                    if (ruleActive.Atom.Type=='Space' or ruleActive.Atom.Type=='Space_Width') and ruleActive.Atom.layer1 == Layer1:
                        if verify_drc_rule(space,ruleActive) == False:
                            ErrorPolygonsList.insert(SpacePolygon.copy())
                        RuleName=ruleActive.Atom.name
                        while ruleActive is not None and ruleActive.Atom.name == RuleName:
                            ruleActive=ruleActive.next
                    else:
                        ruleActive=ruleActive.next
            else:
                SpacePolygon=ext.create_3pnt_polygon(line.pnt1,line.pnt2,p2Active.pnt)
                space=math.sqrt(LineP1PntLength)
                ruleActive=DrcRuleList
                while ruleActive is not None:
                    if (ruleActive.Atom.Type=='Space' or ruleActive.Atom.Type=='Space_Width') and ruleActive.Atom.layer1 == Layer1:
                        if verify_drc_rule(space,ruleActive) == False:
                            ErrorPolygonsList.insert(SpacePolygon.copy())
                        RuleName=ruleActive.Atom.name
                        while ruleActive is not None and ruleActive.Atom.name == RuleName:
                            ruleActive=ruleActive.next
                    else:
                        ruleActive=ruleActive.next
                space=math.sqrt(LineP2PntLength)
                ruleActive=DrcRuleList
                while ruleActive is not None:
                    if (ruleActive.Atom.Type=='Space' or ruleActive.Atom.Type=='Space_Width') and ruleActive.Atom.layer1 == Layer1:
                        if verify_drc_rule(space,ruleActive) == False:
                            ErrorPolygonsList.insert(SpacePolygon.copy())
                        RuleName=ruleActive.Atom.name
                        while ruleActive is not None and ruleActive.Atom.name == RuleName:
                            ruleActive=ruleActive.next
                    else:
                        ruleActive=ruleActive.next
            p2Active=p2Active.next
        p1Active=p1Active.next
    if ErrorPolygonsList.Polygon is not None:
        return ErrorPolygonsList
    else:
        return None

def drc_width(p, DrcRuleList, Layer):
# 
#
# DESCRIPTION: The width of a polygon is calculated using polygon_width and 
# subsequently verified agains a drc rule.
# INPUT: A polygon: p; A rule: DrcRuleList; A text layer: Layer associated with p
#
# OUTPUT: A linked list of polygons
#
# KNOWN ISSUES: The width is only spot checked
#
#
    ErrorPolygonsList=ext.PolygonList()
    width=polygon_width(p)
    ruleActive=DrcRuleList
    while ruleActive is not None:
        if ruleActive.Atom.Type=='Width' and ruleActive.Atom.layer1 == Layer:
            # print('span',width,Layer)
            if verify_drc_rule(width,ruleActive) == False:
                ErrorPolygonsList.insert(ext.create_3pnt_polygon(ext.Point(p.BB.llx,p.BB.lly),ext.Point(p.BB.urx,p.BB.ury),ext.Point(p.BB.llx,p.BB.ury)))
                # print(MinLine.pnt1.x,MinLine.pnt1.y,MinLine.pnt2.x,MinLine.pnt2.y,MinSpanPnt.x,MinSpanPnt.y)
            RuleName=ruleActive.Atom.name
            while ruleActive is not None and ruleActive.Atom.name == RuleName:
                ruleActive=ruleActive.next
        else:
            ruleActive=ruleActive.next
    if ErrorPolygonsList.Polygon is not None:
        return ErrorPolygonsList
    else:
        return None
    
    
def drc_notch(p, DrcRuleList, Layer):
#
# DESCRIPTION: A notch rule check is implemented by spot checking the outside distance
# between two polygon edges.
#
# INPUT: A polygon:p; A rule: DrcRuleList; A text layer associated with p: Layer;
#
# OUTPUT:A linked list of Error Polygons
#
# KNOWN ISSUES: NONE
#
    ErrorPolygonsList=ext.PolygonList()
    pActive=p.head
    midpnt=ext.Point()
    line=ext.Line()
    while pActive.next is not p.head:
        midpnt.x=0.5*(pActive.pnt.x+pActive.next.pnt.x)
        midpnt.y=0.5*(pActive.pnt.y+pActive.next.pnt.y)
        line.pnt1.x=midpnt.x
        line.pnt1.y=midpnt.y
        line.pnt2.x=pActive.next.pnt.x
        line.pnt2.y=pActive.next.pnt.y
        lrot=ext.rotate_line(line,-90)
# Resize line to ensure it covers polygon
        Resize=(p.BB.urx-p.BB.llx+p.BB.ury-p.BB.lly)/math.sqrt((lrot.pnt2.x)**2+(lrot.pnt2.y)**2)
        lrot.pnt2.x=lrot.pnt2.x*Resize
        lrot.pnt2.y=lrot.pnt2.y*Resize
# rotate return a line starting at (0,0)
        lrot.pnt1.x=midpnt.x
        lrot.pnt1.y=midpnt.y
        lrot.pnt2.x=lrot.pnt2.x+midpnt.x
        lrot.pnt2.y=lrot.pnt2.y+midpnt.y
        pnt=ext.find_closest_line_polygon_intersect_pnt(lrot,p)
        if pnt is not None:
            notch=math.sqrt( (lrot.pnt1.x-pnt.x)**2+(lrot.pnt1.y-pnt.y)**2 )
            ruleActive=DrcRuleList
            while ruleActive is not None:
                if ruleActive.Atom.Type=='Notch' and ruleActive.Atom.layer1 == Layer:
                    if verify_drc_rule(notch,ruleActive) == False:
                        ErrorPolygonsList.insert(ext.create_3pnt_polygon(line.pnt1,line.pnt2,pnt))
                    RuleName=ruleActive.Atom.name
                    while ruleActive is not None and ruleActive.Atom.name == RuleName:
                        ruleActive=ruleActive.next
                else:
                    ruleActive=ruleActive.next
        pActive=pActive.next

    if ErrorPolygonsList.Polygon is not None:
        return ErrorPolygonsList
    else:
        return None

def runlength(p1Active, p2Active):
#
# DESCRIPTION: The runlength of two polygon vertex lists is calculated. Only edges
# going in the opposite direction are compared.
#
# INPUT: Two polygon.head lists: p1Active, p2Active
#
# OUTPUT:A floating point number representing the runlength, and the space between
# such edges
#
# KNOWN ISSUES: NONE
#
    pnt1=ext.Point()
    pnt2=ext.Point()
    pnt1.x=p1Active.next.pnt.x-p1Active.pnt.x
    pnt1.y=p1Active.next.pnt.y-p1Active.pnt.y
    pnt2.x=p2Active.next.pnt.x-p2Active.pnt.x
    pnt2.y=p2Active.next.pnt.y-p2Active.pnt.y
    if ext.cross(pnt1,pnt2) < ext.TINY and ext.scalar(pnt1,pnt2) < 0:
        # The edges are parallel and goes in opposite direction
        length1=math.sqrt(pnt1.x**2+pnt1.y**2)
        length2=math.sqrt(pnt2.x**2+pnt2.y**2)
        pnt3=ext.Point()
        pnt3.x=p1Active.pnt.x-p2Active.pnt.x
        pnt3.y=p1Active.pnt.y-p2Active.pnt.y
        runlength=ext.MIN(ext.scalar(pnt2,pnt3)/length2,length2)
        runlength=ext.MIN(runlength,length1)
        space=abs(ext.cross(pnt2,pnt3))/length2
        return runlength,space
    else:
        return ext.MAXV,ext.MAXV

