# -*- coding: utf-8 -*-
"""
Created on Thu Nov  5 11:30:37 2020

@author: msahr
"""
import matplotlib.pyplot as plt
import sys
import re
import math

TINY=1e-10
MINV=-1000000
MAXV=1000000

def MIN(a,b):
    if float(a)<float(b):
        return a
    else:
        return b
    
def MAX(a,b):
    if float(a)>float(b):
        return a
    else:
        return b        

class Cell:
#
# This is the basic definition of a layout cell called: Cell.
# A layout cell is made up of a set of devices and a set of polygons
# THey are represented by two linked lists.
# 
# It consists of two linked lists: plist and DeviceList
# The plist is a list of polygons that belongs to the cell
# and DeviceList is a List of Devices in the Cell
#
# INPUT: self
#
# OUTPUT:A memory pointer to a new polygon
#
# KNOWN ISSUES: NONE
#
    def __init__(self):
#        
# The initialization sets both pointers to None
#        
        self.plist=None
        self.DeviceList=None
    def insert(self,plist):
#        
# This simply sets the plist pointer to the plist argument
# No error checks!! For example, if self.plist != None that 
# list will simply be overwritten
#
# INPUT: a polygon list: plist;
#
# OUTPUT: None
#
# KNOWN ISSUES: None
#
        self.plist=plist
         
class Point:
#
# A point conists of an x, and why coordinate
# This function creates and initializes it to the given
# arguments
#
# INPUT: Two floating point numbers: x, y;
#
# OUTPUT: None
#
# KNOWN ISSUES: None
#
    def __init__(self,x=0, y=0):
        self.x=x
        self.y=y
        
class BoundingBox:
#
# A bounding box is a rectangle that covers a polygon in such a way
# that a point internal to the polygon is also internal to the bounding box.
# This means that a vertex of the polygon could lie on one of the edges (or vertex)
# of the bounding box.
#
# Each polygon is supposed to have a bounding associated with it.
#
# The ll is short for lowerLeft corner and ur is short for upperRight corner of the 
# bounding box
#
    def __init__(self):
        self.llx=MAXV
        self.lly=MAXV
        self.urx=MINV
        self.ury=MINV

def points_equal(pnt1, pnt2):
#
# INPUT: Two points: pnt1, pnt2;
#
# OUTPUT: True/False
#
# KNOWN ISSUES: None
#
    return (pnt1.x==pnt2.x) and (pnt1.y==pnt2.y)

class Line:
# The line class consists of two points that define the beginning and end
# of the line
#
# INPUT: pnt1, pnt2
#
# OUTPUT: A Line element
#
# KNOWN ISSUES: None
#
    def __init__(self):
        self.pnt1=Point()
        self.pnt2=Point()
        
def length_of_line(l):
#
# INPUT: A line
#
# OUTPUT: Floating point with the length of the line
#
# KNOWN ISSUES: None
#
    return math.sqrt( (l.pnt2.x-l.pnt1.x)*(l.pnt2.x-l.pnt1.x)+(l.pnt2.y-l.pnt1.y)*(l.pnt2.y-l.pnt1.y))
        
class Vertex:
# The class Vertex contains a point and a forward and backward pointer. It is used 
# to represent the coordinates of a polygon, forming a closed loop.
#
#
    def __init__(self,x=0,y=0):
        self.pnt=Point(x,y)
        self.next= None
        self.prev= None

class Polygon:
#    
# A polygon consists of four pieces: 
#    1) A pointer to the head of a polygon called: head. This is a pointer to a Vertex
#        Class and the polygon itself is defined by double linked list with 'head' as 
#        first Vertex. The polygons last Vertex points to 'head' and 'head's prev 
#        pointer points to the last Vertex of the polygon
#    2) An integer Nr which equals the number of Vertices in the polygon
#    3) A net pointer where net is an electrical net belonging to Class Net
#    4) A Bounding box, BB, belonging to the BoundingBox Class
#
#
    def __init__(self):
#        
# All entities are initialized to some appropriate value
# 
        self.head= None
        self.Nr=0
        self.net=None
        self.BB=BoundingBox()
    def plot(self,color='#111111'):
# 
# In order to plot a polygon we first create two arrays with the size equal to the
# number of vertices in the polygon self.Nr+1 since we put the first vertex both
# at the beginning and end of the arrays.
#
# INPUT: self and a color
#
# OUTPUT: A plot of all coordinates with lines connecting them in the color
# defined by the input
#
# KNOWN ISSUES: None
#
        x=[0 for i in range(self.Nr+1)]
        y=[0 for i in range(self.Nr+1)]
        vertex=self.head
        i=0
        while (vertex is not None) and (vertex.next is not self.head):
            x[i]=vertex.pnt.x
            y[i]=vertex.pnt.y
            i+=1
            vertex = vertex.next        
        x[self.Nr-1]=vertex.pnt.x
        y[self.Nr-1]=vertex.pnt.y
        x[self.Nr]=self.head.pnt.x
        y[self.Nr]=self.head.pnt.y
        plt.axis('off')
        plt.plot(x,y,color)
    def print(self):
#
# The print statement just goes through the linked list and prints point by point
# of each Vertex
#
# INPUT: self
#
# OUTPUT: All coordinates making up the vertices of the polygon
#
# KNOWN ISSUES: None
#
        vertex = self.head
        while (vertex is not None) and (vertex.next is not self.head):
            print ('(',vertex.pnt.x,',',vertex.pnt.y,')')
            vertex = vertex.next        
        if vertex is not None:
            print ('(',vertex.pnt.x,',',vertex.pnt.y,')')
    def closeLoop(self):
# 
# The closeLoop() function starts at the head vertex and walks through the linked list
# until it hits the last vertex, defined by vertex.next == None. The last vertex
# is double linked to the head vertex by defining the vert.next=self.head and self.head.prev=vertex
#
# INPUT: self
#
# OUTPUT: None
#
# KNOWN ISSUES: None
#
        if self.head is not None:
            if self.head.prev is not None:
                print('Error:[closeLoop] The polygon cannot be closed when using closeLoop')
            vertex=self.head
            Nr=1
            while vertex.next is not None:
                Nr=Nr+1
                vertex=vertex.next
            if vertex.next==None:
                vertex.next=self.head
                self.head.prev=vertex
                self.Nr=Nr
            else:
                print('Error:[insertLast] Last element not found in polygon')
        else:
            print('Error:[closeLoop] Cannot close empty Polygon')
    def insertLast(self,new): 
#
# The insertLast function walks through the polygon linked list antil it finds the last
# vertex, defined by vertx.next==None, and inserts the new vertex after it by defining the
# vertx.next=new and new.prev=vertex
#
# After succesfull Vertex insertion the bounding box needs to be updated
#
# INPUT: self and a new vertex: new;
#
# OUTPUT: None
#
# KNOWN ISSUES: No checks for new is a vertex
#
        if self.head is not None:
            if self.head.prev is not None:
                print('Error:[insertLast] The polygon cannot be closed when using insertLast')
            vertex=self.head
            while vertex.next is not None:
                if vertex == new:
                    print('Error:[instertLast] The polygon already contains this same vertex')
                    sys.exit(1)
                vertex=vertex.next
            if vertex == new:
                print('Error:[instertLast] The polygon already contains this same vertex which is illegal, please review the code')
                sys.exit(1)                
            if vertex.next==None:
                new.next=None
                vertex.next=new
                new.prev=vertex
            else:
                print('Error:[insertLast] Last element not found in polygon')
                sys.exit(1)
        else:
            self.head=new
        self.BB.llx=MIN(self.BB.llx,new.pnt.x)
        self.BB.lly=MIN(self.BB.lly,new.pnt.y)
        self.BB.urx=MAX(self.BB.urx,new.pnt.x)
        self.BB.ury=MAX(self.BB.ury,new.pnt.y)
    def insert(self,start,new): 
#
# The insert statement inserts Vertex new after Vertex start
# The algorithm walks along the polygon looking for the right vertex
# to make sure the vertex start actually is part of the polygon unless
# the polygon is empty.
#
# INPUT: self and two vertices: start and new;
#
# OUTPUT: None
#
# KNOWN ISSUES: No checks for start/new are vertices
#
        if self.head is not None:
            vertex=self.head
            FoundElement=False
            StillInLoop=True
            while StillInLoop and not FoundElement:
                if vertex==start:
                    self.Nr+=1
                    FoundElement=True
                    new.next=vertex.next
                    if new.next is not None:
                        new.next.prev=new
                    vertex.next=new
                    new.prev=vertex
                vertex=vertex.next
                if vertex==self.head:
                    StillInLoop=False
            if not FoundElement:
                print('Error:[insert] start element not found in polygon',vertex.pnt.x,vertex.pnt.y,':',start.pnt.x,start.pnt.y)
                sys.exit(1)
        else:
            self.head=new
        self.BB.llx=MIN(self.BB.llx,new.pnt.x)
        self.BB.lly=MIN(self.BB.lly,new.pnt.y)
        self.BB.urx=MAX(self.BB.urx,new.pnt.x)
        self.BB.ury=MAX(self.BB.ury,new.pnt.y)
    def delete(self,delete):
#        
# The delete function looks for Vertex delete and when found removes it
# from the linked list. If not found it issues an error and exits
#
# When removing a Vertex the bounding box might change. Here the bounding
# box is simply recalculated
#
# INPUT: A vertex to delete: delete;
#
# OUTPUT: None
#
# KNOWN ISSUES: This has not been used, very basic testing done
#
        vertex=self.head
        FoundElement=False
        while (vertex.next is not self.head) and not FoundElement:
            if vertex.next==delete:
                self.Nr-=1
                FoundElement=True
                vertex.next=delete.next
                if delete.next is not None:
                    delete.next.prev=vertex
                delete.prev=None
                delete.next=None
            else:
                vertex=vertex.next
        if not FoundElement:
            print('Error:[delete] start element not found in polygon')
            sys.exit(1)
# Need to redefine the bounding box
        self.BB.llx=MAXV
        self.BB.lly=MAXV
        self.BB.urx=MINV
        self.BB.ury=MINV
        vertex=self.head
        while vertex.next is not self.head:
            self.BB.llx=MIN(self.BB.llx,vertex.pnt.x)
            self.BB.lly=MIN(self.BB.lly,vertex.pnt.y)
            self.BB.urx=MAX(self.BB.urx,vertex.pnt.x)
            self.BB.ury=MAX(self.BB.ury,vertex.pnt.y)
            vertex=vertex.next
# A final check on the last vertex of the polygon
        self.BB.llx=MIN(self.BB.llx,vertex.pnt.x)
        self.BB.lly=MIN(self.BB.lly,vertex.pnt.y)
        self.BB.urx=MAX(self.BB.urx,vertex.pnt.x)
        self.BB.ury=MAX(self.BB.ury,vertex.pnt.y)        
    def copy(self):
#
# The copy function takes a polygon and returns a copy of it
# All vertices are created new in the new polygon so it is a completely new
# entity.
#
# INPUT: self
#
# OUTPUT: return a new polygon
#
# KNOWN ISSUES: No sanity checks are done at all
#
        if self.head==None:
            print('Error: Attempt to copy an empty Polygon')
            exit(1)
        new=Polygon()
        new.Nr=self.Nr
        newVert=Vertex(self.head.pnt.x,self.head.pnt.y)
        new.head=newVert
        origVert=self.head.next
        for i in range(new.Nr-1):
            newVert.next=Vertex(origVert.pnt.x,origVert.pnt.y)
            newVert.next.prev=newVert
            newVert=newVert.next
            origVert=origVert.next
        newVert.next=new.head
        new.head.prev=newVert
        new.BB.llx=self.BB.llx
        new.BB.lly=self.BB.lly
        new.BB.urx=self.BB.urx
        new.BB.ury=self.BB.ury
        return new            
        
        
            
class PolygonList:
#
# The Polygon List is a linked list where each segment points to a polygon
#
    def __init__(self):
        self.Polygon=None
        self.next=None
    def insert(self,p):
#
# DESCRIPTION: The polygon p is inserted last in the list: self;
# INPUT: self, polygon: p;
#
# OUTPUT: None
#
# KNOWN ISSUES: No sanity checks are done
#   
        if self.Polygon != None:
            newp=PolygonList()
            newp.Polygon=p
            pActive=self
            while pActive.next is not None:
                pActive=pActive.next
            pActive.next=newp
        else:
            self.Polygon=p
        
class ListOfPolygonList:
#
# At times it is convenient to have a list of polygonlists. For instance
# in order to represent all layers present in a layout a listOfPolygonlist
# lists each layer as a polygonlist and the list making up all such layers
# is then conveniently described as a list of PolygonLists
# The class ListOfPolygonList is a linked list where each segment points to a 
# PolygonList class. It also has a Name field that is used to label the kind
# of polygons there are, "M0", "PO", "M1" etc.
#
    def __init__(self):
        self.PolyList=None
        self.Name=""
        self.next=None
    def insert(self,plist, Name):
#
# DESCRIPTION: The plist is added last in the self listofpolygonlist
# INPUT: A listofpolygonlist: self; A polygonlist: plist; and a text: Name;
# OUTPUT: None
# KNOWN ISSUES: None
#
        if self.PolyList != None:
            newList=ListOfPolygonList()
            newList.PolyList=plist
            newList.Name=Name
            listActive=self
            while listActive.next is not None:
                listActive=listActive.next
            listActive.next=newList
        else:
            self.PolyList=plist
            self.Name=Name
            
class Device:
#
# The device class contains three pointers to terminals (ports), a device nams:name;
# and some geometry of the device, here: length; width;

    def __init__(self):
        self.term1=None
        self.term2=None
        self.term3=None
        self.name=""
        self.length=-1
        self.width=-1
    
class DeviceList:
#
# The DeviceList class is a linked list where each element points to a class Device
# element and has a forward pointer: next;
#
    def __init__(self):
        self.Device=None
        self.next=None
    def insertDevice(self,Device):
#
# DESCRIPTION: The insertDevice function adds a device last in the devicelist
# if the devicelist points to an empty Device it points to the given Device
#
# INPUT: devicelist: self; Device class: Device;
#
# OUTPUT: None
#
# KNOWN ISSUES: NONE
#
        if Device is not None:
            if self.Device != None:
                dlist=self
                while (dlist.next != None):
                    dlist=dlist.next
                dlist.next=DeviceList()
                dlist=dlist.next
                dlist.Device=Device
            else:
                self.Device=Device
    def cat(self,DeviceList):
#
# DESCRIPTION: Given a DeviceList: self; and a new DeviceList, cat adds the
# DeviceList to the end of self
#
# INPUT: DeviceLists: self, DeviceList;
#
# OUTPUT: None
#
# KNOWN ISSUES: None
#
        while DeviceList is not None:
            self.insertDevice(DeviceList.Device)
            DeviceList=DeviceList.next
    def print(self):
#
# DESCRIPTION: Prints all devices in a devicelist with their terminals and names
# and length/widths
#
# INPUT: self
#
# OUTPUT: text of all devices and their parameters
#
# KNOWN ISSUES: None
# 
        devlist=self
        while devlist is not None:
            print("Device:",devlist.Device.name)
            devlist.Device.term2.print()
            if devlist.Device.length != -1:
                print("Length=",devlist.Device.length)
                print("Width=",devlist.Device.width)
                print("")
            devlist=devlist.next
            
class Net:
#
# DESCRIPTION: A Net class contains aunique identifier for the net with a number
# and a pointer to the next Net
#
    def __init__(self,seed):
        self.netNumber=seed+1
        self.next=None

def scalar(pnt1, pnt2):
#
# DESCRIPTION: Calculates the scalar product between two points
#
# INPUT: Two Point Classes: pnt1, pnt2;
#
# OUTPUT: The scalar product assuming the vectors created by pnt1, pnts
# have their origin at(0,0)
#
# INPUT: Two Point Classes: pnt1, pnt2;
#
# OUTPUT: A floating point value representing the scalar product
#
# KNOWN ISSUES: No sanity checks are done.
#
    return (pnt1.x*pnt2.x+pnt1.y*pnt2.y)    
                
def cross(pnt1, pnt2):
#
# DESCRIPTION: Calculates the cross product between two points
#
# INPUT: Two Point Classes: pnt1, pnt2;
#
# OUTPUT: The cross product assuming the vectors created by pnt1, pnts
# have their origin at(0,0)
#
# KNOWN ISSUES: No sanity checks are done.
#
    return (pnt1.x*pnt2.y-pnt1.y*pnt2.x)

def left(pnt, line):
#
# DESCRIPTION: Ths routine returns true if a pnt lies to the left of the line
#
# INPUT: A class Point: pnt; and a class line:line;
#
# OUTPUT: True/False
#
# KNOWN ISSUES: No sanity checks, like |line|==0 are done
#
    deltaPnt=Point(pnt.x-line.pnt1.x,pnt.y-line.pnt1.y)
    deltaLine=Point(line.pnt2.x-line.pnt1.x,line.pnt2.y-line.pnt1.y)
    return cross(deltaPnt, deltaLine)

def collinear(pnt, line):
#
# DESCRIPTION: This finds out if a pnt and a line are collinear
#
# INPUT: A class point:pnt; a class Line:line;
#
# OUTPUT: True/False
#
# KNOWN ISSUES: No sanity checks, like |line|==0 are done
#
    return abs(left(pnt,line))<TINY

def between(pnt, line):
#
# DESCRIPTION: The routing returns True if a point lies on a line
#
# INPUT: A class point:pnt; a class Line:line;
#
# OUTPUT: True/False
#
# KNOWN ISSUES: No sanity checks, like |line|==0 are done
#
    if collinear(pnt,line) is not True:
        return False
    if line.pnt1.x != line.pnt2.x:
        if (line.pnt1.x < pnt.x) and (line.pnt2.x > pnt.x):
            return True # if the point lies between the line x-pnts it has to also lie between the y-points since it is collinear
        if (line.pnt1.x > pnt.x) and (line.pnt2.x < pnt.x):
            return True
        return False
    else:# the line is vertical
        if (line.pnt1.y < pnt.y) and (line.pnt2.y > pnt.y):
            return True
        if (line.pnt1.y > pnt.y) and (line.pnt2.y < pnt.y):
            return True
        return False
        

def intersect(l1, l2):
#
# DESCRIPTION: This routine returns True if two lines have a proper intersection
#
# INPUT: Two class Line objects: l1, l2;
#
# OUTPUT: True/False
#
# KNOWN ISSUES:No sanity checks, like |line|==0 are done
#
    return left(l1.pnt1,l2)*left(l1.pnt2,l2)<0 and left(l2.pnt1,l1)*left(l2.pnt2,l1)<0

class IntersectList:
#
# DESCRIPTION: The class contains fice elements: A class Point: pnt; 
# a pointer to the next IntersectList object, a variable: used; that is either 
# False/True meaning if the pnt has been used when figuring out a geometrical 
# operation. Lastly, the class contains two pointers to class Vertex: nextp1, 
# nextp2;
#
    def __init__(self):
#
        self.pnt=None
        self.next=None
        self.used=False
        self.nextp1=None
        self.nextp2=None
    def print(self):
#
# DESCRIPTION: This prints out each Vertex point in the IntersectList
#
# INPUT: self
#
# OUTPUT: A text file
#
# KNOWN ISSUES: None
#
        if self.pnt != None:
            print('(',self.pnt.x,self.pnt.y,')')
            element=self.next
            while element is not None:
                print('(',element.pnt.x,element.pnt.y,')')
                element=element.next
    def insert(self,pnt,nextp1,nextp2):
#
# DESCRIPTION: This creates and inserts a new IntersectList point last 
# into list self
#
# INPUT: self, A class Point: pnt; two class Vertex objects: nextp1, nextp2;
#
# OUTPUT: None
#
# KNOWN ISSUES: No sanity checks
#
        if self.pnt != None:
            if self.next is not None:
                element=self.next
                while element.next is not None:
                    element=element.next
            else:
                    element=self
            element.next=IntersectList()
            element=element.next
            element.pnt=Point(pnt.x,pnt.y)
            element.nextp1=nextp1
            element.nextp2=nextp2
        else:
            self.pnt=Point(pnt.x,pnt.y)
            self.nextp1=nextp1
            self.nextp2=nextp2
    def delete(self):
#
# DESCRIPTION: deletes a intersectList object by making all pointers=None
#
# INPUT: self
#
# OUTPUT: None
#
# KNOWN ISSUES: Memory is not released
#
        self.next=None
        self.nextp1=None
        self.nextp2=None
        
def find_line_polygon_intersect_pnt(l, p):
#
# DESCRIPTION: This routine returns the last of the found intersect points between
# a line and a polygon
#
# INPUT: A class Line: l, a class Polygon:p;
#
# OUTPUT: A class Point: isect;
#
# KNOWN ISSUES: No sanity checks
#
    v1=p.head
    l1=Line()
    isect=Point()
    for i in range(p.Nr):
        l1.pnt1=v1.pnt
        l1.pnt2=v1.next.pnt
        if intersect(l1, l):
            print('Found intersect between (',l1.pnt1.x,l1.pnt1.y,'),(',l1.pnt2.x,l1.pnt2.y,') and (',l.pnt1.x,l.pnt1.y,'),(',l.pnt2.x,l.pnt2.y,')')
            if abs(l1.pnt2.x - l1.pnt1.x)>TINY:
                k1=(l1.pnt2.y-l1.pnt1.y)/(l1.pnt2.x-l1.pnt1.x)
                if abs(l.pnt2.x - l.pnt1.x)>TINY:
                    k2=(l.pnt2.y-l.pnt1.y)/(l.pnt2.x-l.pnt1.x)
                    isect.x=(l.pnt1.y-l1.pnt1.y)/(k1-k2)
                    isect.y=l1.pnt1.y+k1*(isect.x-l1.pnt1.x)
                else:
                    isect.x=l.pnt1.x
                    isect.y=l1.pnt1.y+k1*(isect.x-l1.pnt1.x)
            else:
                k2=(l.pnt2.y-l.pnt1.y)/(l.pnt2.x-l.pnt1.x)
                isect.x=l1.pnt1.x
                isect.y=l.pnt1.y+k2*(isect.x-l.pnt1.x)
            print('\t@(',isect.x,isect.y,')')
        v1=v1.next
    return isect

def find_closest_line_polygon_intersect_pnt(l, p):
#
# DESCRIPTION: This routine returns the closest (to l.pnt1) of the found 
# intersect points
#
# INPUT: A class Line: l, a class Polygon:p;
#
# OUTPUT: A class Point: isect;
#
# KNOWN ISSUES: No sanity checks
#
    v1=p.head
    l1=Line()
    isect=Point()
    Closestpnt=Point()
    MinLength=MAXV
    for i in range(p.Nr):
        l1.pnt1=v1.pnt
        l1.pnt2=v1.next.pnt
        if intersect(l1, l):
#            print('Found intersect between (',l1.pnt1.x,l1.pnt1.y,'),(',l1.pnt2.x,l1.pnt2.y,') and (',l.pnt1.x,l.pnt1.y,'),(',l.pnt2.x,l.pnt2.y,')')
            if abs(l1.pnt2.x - l1.pnt1.x)>TINY:
                k1=(l1.pnt2.y-l1.pnt1.y)/(l1.pnt2.x-l1.pnt1.x)
                if abs(l.pnt2.x - l.pnt1.x)>TINY:
                    k2=(l.pnt2.y-l.pnt1.y)/(l.pnt2.x-l.pnt1.x)
                    isect.x=(l.pnt1.y-l1.pnt1.y)/(k1-k2)
                    isect.y=l1.pnt1.y+k1*(isect.x-l1.pnt1.x)
                else:
                    isect.x=l.pnt1.x
                    isect.y=l1.pnt1.y+k1*(isect.x-l1.pnt1.x)
            else:
                k2=(l.pnt2.y-l.pnt1.y)/(l.pnt2.x-l.pnt1.x)
                isect.x=l1.pnt1.x
                isect.y=l.pnt1.y+k2*(isect.x-l.pnt1.x)
#            print('\t@(',isect.x,isect.y,')')
            length=math.sqrt( (isect.x-l.pnt1.x)**2+(isect.y-l.pnt1.y)**2)
            if length < MinLength:
                MinLength=length
                Closestpnt.x=isect.x
                Closestpnt.y=isect.y
        v1=v1.next
    return Closestpnt
        

def find_intersect_pnts(p1, p2):
#
# DESCRIPTION: This routine returns all intersect points between two polygons
# in a class IntersectList linked list
#
# INPUT: Two class Polyon object: p1, p2;
#
# OUTPUT: A class IntersectList: isctlist;
#
# KNOWN ISSUES: No sanity checks
#
    v1=p1.head
    l1=Line()
    l2=Line()
    isect=Point()
    first=True
    isctList=None
    for i in range(p1.Nr):
        l1.pnt1=v1.pnt
        l1.pnt2=v1.next.pnt
        v2=p2.head
        for i in range(p2.Nr):
            l2.pnt1=v2.pnt
            l2.pnt2=v2.next.pnt
            if intersect(l1, l2):
                if first:
                    isctList=IntersectList()
                    first=False
                print('Found intersect between (',l1.pnt1.x,l1.pnt1.y,'),(',l1.pnt2.x,l1.pnt2.y,') and (',l2.pnt1.x,l2.pnt1.y,'),(',l2.pnt2.x,l2.pnt2.y,')')
                if l1.pnt2.x != l1.pnt1.x:
                    k1=(l1.pnt2.y-l1.pnt1.y)/(l1.pnt2.x-l1.pnt1.x)
                    if l2.pnt2.x != l2.pnt1.x:
                        k2=(l2.pnt2.y-l2.pnt1.y)/(l2.pnt2.x-l2.pnt1.x)
                        isect.x=(l2.pnt1.y-l1.pnt1.y)/(k1-k2)
                        isect.y=l1.pnt.y+k1*(isect.x-l1.pnt1.x)
                    else:
                        isect.x=l2.pnt1.x
                        isect.y=l1.pnt1.y+k1*(isect.x-l1.pnt1.x)
                else:
                    k2=(l2.pnt2.y-l2.pnt1.y)/(l2.pnt2.x-l2.pnt1.x)
                    isect.x=l1.pnt1.x
                    isect.y=l2.pnt1.y+k2*(isect.x-l2.pnt1.x)
                print('\t@(',isect.x,isect.y,')')
                isctList.insert(isect,v1.next,v2.next)
            v2=v2.next
        v1=v1.next
    return isctList
        
def insert_intersectpnts_polygon(p,isctList):
#
# DESCRIPTION: All intersect points in a class IntersectList are inserted into
# a polygon. Given an intersect point the routine looks for edges in the polygon 
# on which the intersect point lies.
#
# INPUT: A class Polygon: p; a class IntersectList: isctList
#
# OUTPUT: None
#
# KNOWN ISSUES: Only rudimentary checks if no suitable spot is found, and a
# warning is issued
#
    isct=isctList
    line=Line()
    while isct is not None:
        pnt=isct.pnt
        edge=p.head
        InsertElement=False
        for i in range(p.Nr):
            line.pnt1.x=edge.pnt.x
            line.pnt1.y=edge.pnt.y
            line.pnt2.x=edge.next.pnt.x
            line.pnt2.y=edge.next.pnt.y
            if between(pnt,line):
                InsertElement=True
                v=Vertex(pnt.x,pnt.y)
                p.insert(edge,v)
            edge=edge.next
        if not InsertElement:
            print('WARNING: Could not find a spot to insert:',pnt.x,pnt.y)
        isct=isct.next
        
def delete_intersectpnts_polygon(p):
#
# DESCRIPTION: This routine checks for vertices the lie on an edge between two
# othe vertices. WHen found it is removed from the Vertex list.
#
# INPUT: A class Polygon: p;
#
# OUTPUT: None
#
# KNOWN ISSUES: No sanity checks
#
    line=Line()
    pnt=Point()
    edge=p.head
    TotalVertices=2*p.Nr
    for i in range(TotalVertices):
        line.pnt1.x=edge.pnt.x
        line.pnt1.y=edge.pnt.y
        line.pnt2.x=edge.next.next.pnt.x
        line.pnt2.y=edge.next.next.pnt.y
        pnt.x=edge.next.pnt.x
        pnt.y=edge.next.pnt.y
        if between(pnt,line):
            p.delete(edge.next)
        edge=edge.next
        
def generate_pnt_left(v):
#
# DESCRIPTION: The routine takes a vertex as input and midway along the line
# to the next vertex it puts a point to the left of the edge. It does so by 
# rotating a class Line object 90 degrees counter clockwise
#
# INPUT: class Vertex: v;
#
# OUTPUT: A class Point
#
# KNOWN ISSUES: No sanity check, for example zero line length are done
#
    line=Line()
    line.pnt1=Point(v.pnt.x,v.pnt.y)
    line.pnt2=Point(v.next.pnt.x,v.next.pnt.y)
    midpnt=Point( (v.pnt.x+v.next.pnt.x)*0.5, (v.pnt.y+v.next.pnt.y)*0.5)
    lrot=rotate_line(line,90)
    leftpnt=Point()
    leftpnt.x=midpnt.x+lrot.pnt2.x
    leftpnt.y=midpnt.y+lrot.pnt2.y
    return leftpnt

def generate_pnt_right(v):
#
# DESCRIPTION: The routine takes a vertex as input and midway along the line
# to the next vertex it puts a point to the right of the edge. It does so by 
# rotating a class Line object 90 degrees clockwise
#
# INPUT: class Vertex: v;
#
# OUTPUT: A class Point
#
# KNOWN ISSUES: No sanity check, for example zero line length are done
#
    line=Line()
    line.pnt1=Point(v.pnt.x,v.pnt.y)
    line.pnt2=Point(v.next.pnt.x,v.next.pnt.y)
    midpnt=Point( (v.pnt.x+v.next.pnt.x)*0.5, (v.pnt.y+v.next.pnt.y)*0.5)
    lrot=rotate_line(line,-90)
    lrot_length=length_of_line(lrot)
    lrot.pnt2.x=lrot.pnt2.x/(lrot_length*1e4)
    lrot.pnt2.y=lrot.pnt2.y/(lrot_length*1e4)    
    rightpnt=Point()
    rightpnt.x=midpnt.x+lrot.pnt2.x
    rightpnt.y=midpnt.y+lrot.pnt2.y
    return rightpnt

def rotate_line(l,angle):
# 
# DESCRIPTION: A class Line is rotated an angle: angle; by moving the line
# so the line.pnt1=(0,0)
#
# INPUT: A class Line: l;  A floating point: angle;
#
# OUTPUT: A new class Line; newl;
#
# KNOWN ISSUES: No sanity checks, for instance check for |l|=0 are done
#
    newl=Line()
    if l.pnt1.x != 0 or l.pnt1.y != 0:
        newl.pnt2.x=l.pnt2.x-l.pnt1.x
        newl.pnt2.y=l.pnt2.y-l.pnt1.y
    oldx=newl.pnt2.x
    oldy=newl.pnt2.y
    newl.pnt2.x=math.cos(angle*math.pi/180)*oldx-math.sin(angle*math.pi/180)*oldy
    newl.pnt2.y=math.sin(angle*math.pi/180)*oldx+math.cos(angle*math.pi/180)*oldy
    
    return newl
    
    
    
        
def find_vertex_pnt(p,pnt):
#
# DESCRIPTION: This routine looks for a vertex in a polygon with coordinates
# given by a class Point object
#
# INPUT: class Polygon: p; class Point: pnt;
#
# OUTPUT: class vertex: None if not found
#
# KNOWN ISSUES: No sanity checks, if no vertex is found a simple warning is issued
#
    vertex=p.head
    StillInLoop=True
    FoundVertex=False
    while StillInLoop and not FoundVertex:
        if (vertex.pnt.x == pnt.x) and (vertex.pnt.y == pnt.y) :
            FoundVertex=True
            return vertex
        vertex=vertex.next
        if vertex == p.head:
            StillInLoop=False
    if not FoundVertex:
        print('Warning: vertex point not found in Polygon')
    return None

def pnt_in_intersect_list(pnt,isctList, MarkUsed=False):
#
# DESCRIPTION: This routine walks through an IntersectList object and looks
# for an specific coordinate given by  class Point object. If found it updates
# the IntersectList object used field as True if a boolean input is True (if 
# boolean input is False nothing is done), it returns True. If not found it
# returns False  
#
# INPUT: A class Point: pnt; a class IntersectList: isctList; a boolean: MarkUsed
#
# OUTPUT:True/False depending on if coordinates given by pnt are found
#
# KNOWN ISSUES: If MarkUsed=False the isct.used field is not updated. 
# No sanity checks
#
    isct=isctList
    while isct is not None:
        if isct.pnt.x == pnt.x and isct.pnt.y==pnt.y:
            if MarkUsed:
                isct.used=True
            return True
        isct=isct.next
    return False

def clear_used_in_intersect_list(isctList):
#
# DESCRIPTION: THis routing updates all IntersectList elements field used to False
#
# INPUT: A class IntersectList: isctList;
#
# OUTPUT: None
#
# KNOWN ISSUES: No sanity checks, it isctList=None no warning issued
#
    isct=isctList
    while isct is not None:
        isct.used=False
        isct=isct.next
        
def still_unused_intersect_pnts(isctList):
#
# DESCRIPTION: This routine loops through an IntersetList list and returns True
# if an IntersetList element has a field used=False. If not found it returns
# False
#
# INPUT: An IntersectList: isctList;
#
# OUTPUT: True/False
#
# KNOWN ISSUES: No sanity checks, it isctList=None no warning issued
#
    isct=isctList
    while isct is not None:
        if isct.used==False:
            return True
        isct=isct.next
    return False

def find_unused_intersect_pnt(isctList):
#
# DESCRIPTION: This routine loops through an IntersectList linked list and
# return True if an used=False IntersectList element is found.
#
# INPUT: A class IntersectList: isctList;
#
# OUTPUT: True/False
#
# KNOWN ISSUES: No sanity checks, it isctList=None no warning issued
#
    isct=isctList
    while isct is not None:
        if isct.used==False:
            return isct.pnt
        isct=isct.next
    return None

def point_inside_polygon(pnt, p):
#
# DESCRIPTION: This rouitine returns True if a class Point: pnt; is inside a 
# polygon by creating a line from p to a pnt 'guaranteed' to be outside the
# polygon. It then counts the number of time the line intersect the Polygon
# If that count is an odd number the point resides inside the polygon
#
# INPUT: class Point: pnt; class Polygon: p;
#
# OUTPUT: True/False
#
# KNOWN ISSUES: The guaranteed outside point is just a large number. It should
# use the polygons bounding box to establish the outside point.
#
    l1=Line()
    l2=Line()
    l1.pnt1.x=pnt.x
    l1.pnt1.y=pnt.y
    l1.pnt2.x=-100000000 # should really use bounding box to set the limit
    l1.pnt2.y=pnt.y
    v=p.head
    NIntersect=0
    for i in range(p.Nr):
        l2.pnt1=v.pnt
        l2.pnt2=v.next.pnt
        if intersect(l1,l2): # This does not account for line crossing vertices in p
            NIntersect+=1
        v=v.next
    return NIntersect/2 != int(NIntersect/2)

def polygon_inside_polygon(p1, p2):
#
#
# DESCRIPTION: This rouitine returns True if one vertex of one of two input 
# polygons is found inside the other polygon. 
#
# INPUT: Two class Polygon: p1, p2;
#
# OUTPUT: True/False
#
# KNOWN ISSUES: No sanity checks
#
    if point_inside_polygon(p1.head.pnt, p2) or point_inside_polygon(p2.head.pnt, p1):
        return True

def poly_touch(p1, p2):
# 
# DESCRIPTION: This routine returns True if either the input polygons are found
# inside each other OR if a vertex lies on one of the other polygons edges.
#
# INPUT: Two class Polygon: p1, p2;
#
# OUTPUT: True/False
#
# KNOWN ISSUES: No sanity checks
#
    v1=p1.head
    l1=Line()
    l2=Line()
    if polygon_inside_polygon(p1,p2): 
        return True
    for i in range(p1.Nr):
        l1.pnt1=v1.pnt
        l1.pnt2=v1.next.pnt
        v2=p2.head
        for i in range(p2.Nr):
            l2.pnt1=v2.pnt
            l2.pnt2=v2.next.pnt
            if intersect(l1, l2):
                return True
            if between(l1.pnt1,l2) or between(l1.pnt2,l2) or between(l2.pnt1,l1) or between(l2.pnt2,l1):
                return True
            if points_equal(l1.pnt1,l2.pnt1) or points_equal(l1.pnt2,l2.pnt1) or points_equal(l1.pnt1,l2.pnt2) or points_equal(l1.pnt2,l2.pnt2):
                return True
            v2=v2.next
        v1=v1.next
    return False
        

def poly_AND_NOT(p1in, p2in):
#
# DESCRIPTION: This routine computes an AND NOT operation of two polygons,
# assuming only proper intersections are to be found
#
# INPUT: Two Polygons
#
# OUTPUT: A linked ListOfPolygons: PolyList
#
# KNOWN ISSUES:No error checks on input to make sure polygons are well behaved
#              No Checks for improper intersections 
#              No Checks for single intersect point 
#
#    First the input polygons are copied to local polygons. This is needed since
#    the polygons will be edited with the intersect point list
    p1=p1in.copy()
    p2=p2in.copy()
    isctList=find_intersect_pnts(p1,p2)
    insert_intersectpnts_polygon(p1,isctList)
    insert_intersectpnts_polygon(p2,isctList)
    PolyList=PolygonList()
#
#    For the case where there are no intersect points either the polygons are totally
#    separate or one is inside the other. In the former case, the result is None. In
#    the latter case the result is the polygon that is inside the other
#
    if isctList==None:
        if point_inside_polygon(p1.head.pnt,p2):
            pANDNOT=None
        elif point_inside_polygon(p2.head.pnt,p1):
            print('Error: Illegal Polygon operation. The NOT polygon lies inside the first polygon')
            print('Causing a donut as a result. This is not allowed in this implementation.')
            p1.print()
            p2.print()
            exit(1)
        return None
    clear_used_in_intersect_list(isctList)
#    
#    Start by walking through the resulting polygon by picking one of the intersect
#    points that are not used.
#    
    while still_unused_intersect_pnts(isctList):
        startPnt=find_unused_intersect_pnt(isctList)
        v1=find_vertex_pnt(p1,startPnt)
        v2=find_vertex_pnt(p2,startPnt)
        l1=Line()
        StillInLoop=True
        new=Vertex(startPnt.x,startPnt.y)
        pANDNOT=Polygon()
        pANDNOT.insert(pANDNOT.head,new)
        isctList.used=True
        while StillInLoop:
            l1.pnt1.x=v1.pnt.x
            l1.pnt1.y=v1.pnt.y
            l1.pnt2.x=v1.next.pnt.x
            l1.pnt2.y=v1.next.pnt.y
#            
#            Find out which polygon to use for the next vertex by looking at the 
#            next point and determining if it is to the left of l1, which points along
#            v1, or not
#
            if left(v2.prev.pnt,l1)<0 :
                print('choosing poly2')
                ReverseTravel=True
                pRunner=v2
            else:
                print('choosing poly1')
                ReverseTravel=False
                pRunner=v1
            OnPolygon=True
#            
#            We now walk along the chosen polygon until the next intersect point is found
#
            while OnPolygon:
                if ReverseTravel:
                    pRunner=pRunner.prev
                else:
                    pRunner=pRunner.next
                if pnt_in_intersect_list(pRunner.pnt,isctList, True):
                    OnPolygon=False
#                    
#                    If we have encountered the starting point it means we have found the full
#                    polygon and we can close it and continue by looking for unused
#                    interssect points. If not we need to determine which polygon to walk
#                    along next  
# 
                    if (pRunner.pnt.x == startPnt.x) and (pRunner.pnt.y==startPnt.y):
                        StillInLoop=False
                        pANDNOT.closeLoop()
                    else:
                        new=Vertex(pRunner.pnt.x,pRunner.pnt.y)
                        print('adding isectpnt',new.pnt.x,new.pnt.y)
                        pANDNOT.insertLast(new)
                        v1=find_vertex_pnt(p1,pRunner.pnt)
                        v2=find_vertex_pnt(p2,pRunner.pnt)
                else:
                    new=Vertex(pRunner.pnt.x,pRunner.pnt.y)
                    print('adding poly pnt',new.pnt.x,new.pnt.y)
                    pANDNOT.insertLast(new)
#        
#        The polygon is added to the list of polygons and we continue to look for the
#        next unsed intersect point
# 
        PolyList.insert(pANDNOT)
    return PolyList


def poly_AND(p1in, p2in):
#
# DESCRIPTION: This routine computes an AND operation of two polygons,
# assuming only proper intersections are to be found
#
# INPUT: Two Polygons
#
# OUTPUT: A linked ListOfPolygons: PolyList
#
# KNOWN ISSUES:No error checks on input to make sure polygons are well behaved
#              No Checks for improper intersections 
#              No Checks for single intersect point 
#
#    First the input polygons are copied to local polygons. This is needed since
#    the polygons will be edited with the intersect point list
    
    p1=p1in.copy()
    p2=p2in.copy()
    isctList=find_intersect_pnts(p1,p2)
    insert_intersectpnts_polygon(p1,isctList)
    insert_intersectpnts_polygon(p2,isctList)
    PolyList=PolygonList()
#
#    For the case where there are no intersect points either the polygons are totally
#    separate or one is inside the other. In the former case, the result is None. In
#    the latter case the result is the polygon that is inside the other
#    
    if isctList==None:
        pAND=None
        if point_inside_polygon(p1.head.pnt,p2):
            pAND=p1.copy()
        elif point_inside_polygon(p2.head.pnt,p1):
            pAND=p2.copy()
        if pAND==None:
            return None
        PolyList.insert(pAND)
    clear_used_in_intersect_list(isctList)
#    
#    Start by walking through the resulting polygon by picking one of the intersect
#    points that are not used.
#    
    while still_unused_intersect_pnts(isctList):
        startPnt=find_unused_intersect_pnt(isctList)
        v1=find_vertex_pnt(p1,startPnt)
        v2=find_vertex_pnt(p2,startPnt)
        l1=Line()
        StillInLoop=True
        new=Vertex(startPnt.x,startPnt.y)
        pAND=Polygon()
        pAND.insert(pAND.head,new)
        isctList.used=True
        print('first:',new.pnt.x,new.pnt.y)
        while StillInLoop:
            l1.pnt1.x=v1.pnt.x
            l1.pnt1.y=v1.pnt.y
            l1.pnt2.x=v1.next.pnt.x
            l1.pnt2.y=v1.next.pnt.y
#            print('left:',left(v2.next.pnt,l1))
#            
#            Find out which polygon to use for the next vertex by looking at the 
#            next point and determining if it is to the left of l1, which points along
#            v1, or not
#            
            if left(v2.next.pnt,l1)<0 :
                print('choosing poly2')
                pRunner=v2
            else:
                print('choosing poly1')
                pRunner=v1
            OnPolygon=True
#            
#            We now walk along the chosen polygon until the next intersect point is found
#            
            while OnPolygon:
                pRunner=pRunner.next
                if pnt_in_intersect_list(pRunner.pnt,isctList, True):
                    OnPolygon=False
#                    
#                    If we have encountered the starting point it means we have found the full
#                    polygon and we can close it and continue by looking for unused
#                    interssect points. If not we need to determine which polygon to walk
#                    along next  
#                    
                    if (pRunner.pnt.x == startPnt.x) and (pRunner.pnt.y==startPnt.y):
                        StillInLoop=False
                        pAND.closeLoop()
                    else:
                        new=Vertex(pRunner.pnt.x,pRunner.pnt.y)
                        print('adding isectpnt',new.pnt.x,new.pnt.y)
                        pAND.insertLast(new)
                        v1=find_vertex_pnt(p1,pRunner.pnt)
                        v2=find_vertex_pnt(p2,pRunner.pnt)
                else:
                    new=Vertex(pRunner.pnt.x,pRunner.pnt.y)
                    print('adding poly pnt',new.pnt.x,new.pnt.y)
                    pAND.insertLast(new)
#        
#        The polygon is added to the list of polygons and we continue to look for the
#        next unsed intersect point
#        
        PolyList.insert(pAND)
    return PolyList
         
def poly_OR(p1in, p2in):
#
# DESCRIPTION: This routine computes an OR operation of two polygons,
# assuming only proper intersections are to be found
#
# INPUT: Two Polygons
#
# OUTPUT: A linked ListOfPolygons: PolyList
#
# KNOWN ISSUES:No error checks on input to make sure polygons are well behaved
#              No Checks for improper intersections 
#              No Checks for single intersect point 
#
#    First the input polygons are copied to local polygons. This is needed since
#    the polygons will be edited with the intersect point list
    p1=p1in.copy()
    p2=p2in.copy()
    isctList=find_intersect_pnts(p1,p2)
    insert_intersectpnts_polygon(p1,isctList)
    insert_intersectpnts_polygon(p2,isctList)
    PolyList=PolygonList()
#
#    For the case where there are no intersect points either the polygons are totally
#    separate or one is inside the other. In the former case, the result is None. In
#    the latter case the result is the polygon that is inside the other
#
    if isctList==None:
        pOR=None
        if point_inside_polygon(p1.head.pnt,p2):
            pOR=p2.copy()
        elif point_inside_polygon(p2.head.pnt.p1):
            pOR=p1.copy()
        if pOR==None:
            PolyList.insert(p1)
            PolyList.insert(p2)
        PolyList.insert(pOR)
        return PolyList

    clear_used_in_intersect_list(isctList)
#    
#    Start by walking through the resulting polygon by picking one of the intersect
#    points that are not used.
#     
    startPnt=isctList.pnt
    v1=find_vertex_pnt(p1,startPnt)
    v2=find_vertex_pnt(p2,startPnt)
    l1=Line()
    StillInLoop=True
    new=Vertex(startPnt.x,startPnt.y)
    pOR=Polygon()
    pOR.insert(pOR.head,new)
    while StillInLoop:
        l1.pnt1.x=v1.pnt.x
        l1.pnt1.y=v1.pnt.y
        l1.pnt2.x=v1.next.pnt.x
        l1.pnt2.y=v1.next.pnt.y
#            
#            Find out which polygon to use for the next vertex by looking at the 
#            next point and determining if it is to the left of l1, which points along
#            v1, or not
#
        if left(v2.next.pnt,l1)>0 :
            print('choosing poly2')
            pRunner=v2
        else:
            print('choosing poly1')
            pRunner=v1
        OnPolygon=True
#            
#            We now walk along the chosen polygon until the next intersect point is found
#
        while OnPolygon:
            pRunner=pRunner.next
            if pnt_in_intersect_list(pRunner.pnt,isctList):
                OnPolygon=False
#                    
#                    If we have encountered the starting point it means we have found the full
#                    polygon and we can close it and continue by looking for unused
#                    interssect points. If not we need to determine which polygon to walk
#                    along next  
#
                if (pRunner.pnt.x == startPnt.x) and (pRunner.pnt.y==startPnt.y):
                    StillInLoop=False
                    pOR.closeLoop()
                else:
                    new=Vertex(pRunner.pnt.x,pRunner.pnt.y)
                    print('adding isectpnt',new.pnt.x,new.pnt.y)
                    pOR.insertLast(new)
                    v1=find_vertex_pnt(p1,pRunner.pnt)
                    v2=find_vertex_pnt(p2,pRunner.pnt)
            else:
                new=Vertex(pRunner.pnt.x,pRunner.pnt.y)
                print('adding poly pnt',new.pnt.x,new.pnt.y)
                pOR.insertLast(new)
#        
#        The polygon is added to the list of polygons and we continue to look for the
#        next unsed intersect point
# 
    return pOR

def polygon_area(p):
#
# DESCRIPTION: This routine returns the area of a given polygon
#
# INPUT: class Polhyon: p;
#
# OUTPUT: floating point number
#
# KNOWN ISSUES: No sanity checks
#
    Area=0
    vertex = p.head
    while (vertex is not None) and (vertex.next is not p.head):
        Area+=vertex.pnt.x*vertex.next.pnt.y-vertex.pnt.y*vertex.next.pnt.x
        vertex = vertex.next        
    Area+=p.head.prev.pnt.x*p.head.pnt.y-p.head.prev.pnt.y*p.head.pnt.x
    return Area*0.5

def polygon_circumference(p):
#
# DESCRIPTION: This routine returns the circumference of a given polygon
#
# INPUT: class Polhyon: p;
#
# OUTPUT: floating point number
#
# KNOWN ISSUES: No sanity checks
#
    Circumference=0
    vertex = p.head
    while (vertex is not None) and (vertex.next is not p.head):
        Circumference+=math.sqrt((vertex.pnt.x-vertex.next.pnt.x)**2+(vertex.pnt.y-vertex.next.pnt.y)**2)
        vertex = vertex.next        
    Circumference+=math.sqrt((p.head.pnt.x-p.head.prev.pnt.x)**2+(p.head.pnt.y-p.head.prev.pnt.y)**2)
    return Circumference

def create_3pnt_polygon(p1,p2,p3):
#
# DESCRIPTION: a class Polygon object is created from three input class Point
# objects
#
# INPUT: Three class Points: p1, p2, p3;
#
# OUTPUT: A class Polygon element
#
# KNOWN ISSUES: No sanity checks on the input points
#
    p=Polygon()
    v1=Vertex(p1.x,p1.y)
    v2=Vertex(p2.x,p2.y)
    v3=Vertex(p3.x,p3.y)
    p.insert(p.head,v1)
    p.insertLast(v2)
    p.insertLast(v3)
    p.closeLoop()
    return p
            
def read_layout(layout):
#
# DESCRIPTION: This routine parses an input file and returns a class
# ListOfPolygonList object with all polygons found
#
# INPUT: An input file
#
# OUTPUT: A class ListPolygonList linked list
#
# KNOWN ISSUES: Few error checks are returned. The routine is very sensitive
# to spaces and syntax
#
    try:
        myfile=open(layout,'r')
    except:
        print('layout file',layout,' not found')
        sys.exit()
    LayerCount=0
    PolygonON=False
    line=myfile.readline()
    ListOfPList=ListOfPolygonList()
    while line !='END\n' :
        if re.split(' ',line)[0]=='Layer:':
            LayerCount+=1
            ListOfPolygons=PolygonList()
            Name=re.split('[ \n]',line)[1]
            ListOfPList.insert(ListOfPolygons, Name)
        elif line[0]=='(':
            if not PolygonON:
                PolygonON=True
                p=Polygon()
                ListOfPolygons.insert(p)
            v=Vertex()
            v.pnt.x=float(re.split(' |\(|,|\)',line)[1])
            v.pnt.y=float(re.split(' |\(|,|\)',line)[2])
            p.insertLast(v)
        elif line[0]=='\n' and PolygonON:
            PolygonON=False
            p.closeLoop()      
        line=myfile.readline()                  
    return ListOfPList

def find_ListOfPolygonList(Cell,name):
#
# DESCRIPTION: This routine searches a class Cell object for a ListOfPOlygonList
# element that matches a certain text name. If found it the returns the list
# and of not found it returns None.
#
# INPUT: class Cell: Cell; text: name;
#
# OUTPUT: a linked ListOfPolygonlist, or None
#
# KNOWN ISSUES: No sanity checks
#
    plist=Cell.plist
    while (plist != None):
        if plist.Name==name:
            return plist.PolyList
        else:
            plist=plist.next
    return None

def find_nmos_body(Cell):
#
# DESCRIPTION: This routine looks for ListOfPolygonLists with name
# "Poly", "Diffusion" and "NP". If they are found it creates nmos bodies
# performing AND operations between these three lists of polygons
#
# INPUT: class Cell: Cell;
#
# OUTPUT: A class ListofPolygonList named "NMOSbody" containing all found nmos
# bodies.
#
# KNOWN ISSUES: No sanity checks
#
    ListOfNMOSBody=PolygonList()
    poly=find_ListOfPolygonList(Cell,"Poly")
    while poly is not None:
        diff=find_ListOfPolygonList(Cell,"Diffusion")
        while diff is not None:
            ppolydiff=poly_AND(poly.Polygon,diff.Polygon)
            delete_intersectpnts_polygon(poly.Polygon)
            delete_intersectpnts_polygon(diff.Polygon)            
            nplus=find_ListOfPolygonList(Cell, "NP")
            while nplus is not None:
                while ppolydiff is not None:
                    nmosbody=poly_AND(ppolydiff.Polygon,nplus.Polygon)
                    delete_intersectpnts_polygon(ppolydiff.Polygon)
                    delete_intersectpnts_polygon(nplus.Polygon)
                    while nmosbody != None:
                        ListOfNMOSBody.insert(nmosbody.Polygon)
                        nmosbody=nmosbody.next
                    ppolydiff=ppolydiff.next
                nplus=nplus.next
            diff=diff.next
        poly=poly.next
    if ListOfNMOSBody.Polygon != None:
        Cell.plist.insert(ListOfNMOSBody,'NMOSbody')                            
    return 

def find_nmos_terminals(Cell):
#
# DESCRIPTION: This routine looks for ListOfPolygonLists with name
# "Poly", "Diffusion" and "NP". If they are found it creates nmos bodies
# performing AND, and AND-NOT operations between these three lists of polygons
#
# INPUT: class Cell: Cell;
#
# OUTPUT: A class ListofPolygonList named "NMOSTerminal" containing all found nmos
# bodies.
#
# KNOWN ISSUES: No sanity checks
#
    ListOfNMOSTerminals=PolygonList()
    poly=find_ListOfPolygonList(Cell,"Poly")
    while poly is not None:
        diff=find_ListOfPolygonList(Cell,"Diffusion")
        while diff is not None:
            pdiffNOTpoly=poly_AND_NOT(diff.Polygon,poly.Polygon)
            delete_intersectpnts_polygon(poly.Polygon)
            delete_intersectpnts_polygon(diff.Polygon)            
            nplus=find_ListOfPolygonList(Cell, "NP")
            while nplus is not None:
                while pdiffNOTpoly is not None:
                    nmosterminal=poly_AND(pdiffNOTpoly.Polygon,nplus.Polygon)
                    delete_intersectpnts_polygon(pdiffNOTpoly.Polygon)
                    delete_intersectpnts_polygon(nplus.Polygon)
                    while nmosterminal != None:
                        ListOfNMOSTerminals.insert(nmosterminal.Polygon)
                        nmosterminal=nmosterminal.next
                    pdiffNOTpoly=pdiffNOTpoly.next
                nplus=nplus.next
            diff=diff.next
        poly=poly.next
    if ListOfNMOSTerminals.Polygon != None:
        Cell.plist.insert(ListOfNMOSTerminals,'NMOSTerminal')                            
    return 

def find_nmos_devices(Cell):
#
# DESCRIPTION: This routine looks for ListOfPolygonLists with name
# "NMOSbody" and "NMOSTerminal". If they are found it creates "NMOS" devices 
# by checking for NMOSTerminals touching NMOSbody polygons. Both source and 
# drain needs a separate NMOSTerminal polygon. It also computes the channel 
# length and width by checking on which edge of the transistor body polygon
# the drain/source polygons lie.
#
# INPUT: class Cell: Cell;
#
# OUTPUT: A class Device list named "NMOS" containing all found nmos devices
#
# KNOWN ISSUES: No sanity checks. The routine implicitly assumes all polygons
# are rectangles
#
    ListOfNMOSBody=find_ListOfPolygonList(Cell,'NMOSbody')
    ListOfNMOSTerminal=find_ListOfPolygonList(Cell,'NMOSTerminal')
    if ListOfNMOSBody==None or ListOfNMOSTerminal==None :
        return None
    plist=ListOfNMOSBody
    DevList=DeviceList()
    DevList.Device=Device()
    DevList.Device.name="NMOS"
    DevListActive=DevList
    while (plist != None):
        DevListActive.Device.term2=plist.Polygon
        TermList=ListOfNMOSTerminal
        Done=False
        while (TermList != None and not Done):
            if poly_touch(plist.Polygon,TermList.Polygon):
                if DevListActive.Device.term1 == None:
                    DevListActive.Device.term1=TermList.Polygon
                else:
                    if DevListActive.Device.term1 != TermList.Polygon:
                        DevListActive.Device.term3=TermList.Polygon
                        Done=True
                        pntright=generate_pnt_right(DevListActive.Device.term2.head)
                        if point_inside_polygon(pntright, DevListActive.Device.term1) or point_inside_polygon(pntright,DevListActive.Device.term3):
                            DevListActive.Device.width=abs(DevListActive.Device.term2.head.pnt.x-DevListActive.Device.term2.head.next.pnt.x)
                            DevListActive.Device.width+=abs(DevListActive.Device.term2.head.pnt.y-DevListActive.Device.term2.head.next.pnt.y)
                            DevListActive.Device.length=abs(DevListActive.Device.term2.head.prev.pnt.x-DevListActive.Device.term2.head.pnt.x)
                            DevListActive.Device.length+=abs(DevListActive.Device.term2.head.prev.pnt.y-DevListActive.Device.term2.head.pnt.y)
                        else:
                            DevListActive.Device.length=abs(DevListActive.Device.term2.head.pnt.x-DevListActive.Device.term2.head.next.pnt.x)
                            DevListActive.Device.length+=abs(DevListActive.Device.term2.head.pnt.y-DevListActive.Device.term2.head.next.pnt.y)
                            DevListActive.Device.width=abs(DevListActive.Device.term2.head.prev.pnt.x-DevListActive.Device.term2.head.pnt.x)
                            DevListActive.Device.width+=abs(DevListActive.Device.term2.head.prev.pnt.y-DevListActive.Device.term2.head.pnt.y)                                                        
            TermList=TermList.next
        plist=plist.next
        if plist is not None:
            DevListActive.next=DeviceList()
            DevListActive=DevListActive.next
            DevListActive.Device=Device()
            DevListActive.Device.name="NMOS"
            

        
    return DevList

def find_pmos_body(Cell):
#
# DESCRIPTION: This routine looks for ListOfPolygonLists with name
# "Poly", "Diffusion" and "PP". If they are found it creates pmos bodies
# performing AND operations between these three lists of polygons
#
# INPUT: class Cell: Cell;
#
# OUTPUT: A class ListofPolygonList named "PMOSbody" containing all found pmos
# bodies.
#
# KNOWN ISSUES: No sanity checks
#
    ListOfPMOSBody=PolygonList()
    poly=find_ListOfPolygonList(Cell,"Poly")
    while poly is not None:
        diff=find_ListOfPolygonList(Cell,"Diffusion")
        while diff is not None:
            ppolydiff=poly_AND(poly.Polygon,diff.Polygon)
            delete_intersectpnts_polygon(poly.Polygon)
            delete_intersectpnts_polygon(diff.Polygon)            
            pplus=find_ListOfPolygonList(Cell, "PP")
            while pplus is not None:
                while ppolydiff is not None:
                    pmosbody=poly_AND(ppolydiff.Polygon,pplus.Polygon)
                    delete_intersectpnts_polygon(ppolydiff.Polygon)
                    delete_intersectpnts_polygon(pplus.Polygon)
                    while pmosbody != None:
                        ListOfPMOSBody.insert(pmosbody.Polygon)
                        pmosbody=pmosbody.next
                    ppolydiff=ppolydiff.next
                pplus=pplus.next
            diff=diff.next
        poly=poly.next
    if ListOfPMOSBody.Polygon != None:
        Cell.plist.insert(ListOfPMOSBody,'PMOSbody')                            
    return 

def find_pmos_terminals(Cell):
#
# DESCRIPTION: This routine looks for ListOfPolygonLists with name
# "Poly", "Diffusion" and "PP". If they are found it creates pmos bodies
# performing AND, and AND-NOT operations between these three lists of polygons
#
# INPUT: class Cell: Cell;
#
# OUTPUT: A class ListofPolygonList named "PMOSTerminal" containing all found pmos
# bodies.
#
# KNOWN ISSUES: No sanity checks
#
    ListOfPMOSTerminals=PolygonList()
    poly=find_ListOfPolygonList(Cell,"Poly")
    while poly is not None:
        diff=find_ListOfPolygonList(Cell,"Diffusion")
        while diff is not None:
            pdiffNOTpoly=poly_AND_NOT(diff.Polygon,poly.Polygon)
            delete_intersectpnts_polygon(poly.Polygon)
            delete_intersectpnts_polygon(diff.Polygon)            
            pplus=find_ListOfPolygonList(Cell, "PP")
            while pplus is not None:
                while pdiffNOTpoly is not None:
                    pmosterminal=poly_AND(pdiffNOTpoly.Polygon,pplus.Polygon)
                    delete_intersectpnts_polygon(pdiffNOTpoly.Polygon)
                    delete_intersectpnts_polygon(pplus.Polygon)
                    while pmosterminal != None:
                        ListOfPMOSTerminals.insert(pmosterminal.Polygon)
                        pmosterminal=pmosterminal.next
                    pdiffNOTpoly=pdiffNOTpoly.next
                pplus=pplus.next
            diff=diff.next
        poly=poly.next
    if ListOfPMOSTerminals.Polygon != None:
        Cell.plist.insert(ListOfPMOSTerminals,'PMOSTerminal')                            
    return 

def find_pmos_devices(Cell):
#
# DESCRIPTION: This routine looks for ListOfPolygonLists with name
# "PMOSbody" and "PMOSTerminal". If they are found it creates "PMOS" devices 
# by checking for PMOSTerminals touching NMOSbody polygons. Both source and 
# drain needs a separate PMOSTerminal polygon. It also computes the channel 
# length and width by checking on which edge of the transistor body polygon
# the drain/source polygons lie.
#
# INPUT: class Cell: Cell;
#
# OUTPUT: A class Device list named "PMOS" containing all found nmos devices
#
# KNOWN ISSUES: No sanity checks. The routine implicitly assumes all polygons
# are rectangles
#
    ListOfPMOSBody=find_ListOfPolygonList(Cell,'PMOSbody')
    ListOfPMOSTerminal=find_ListOfPolygonList(Cell,'PMOSTerminal')
    if ListOfPMOSBody==None or ListOfPMOSTerminal==None :
        return None
    plist=ListOfPMOSBody
    DevList=DeviceList()
    DevList.Device=Device()
    DevList.Device.name="PMOS"
    DevListActive=DevList
    while (plist != None):
        DevListActive.Device.term2=plist.Polygon
        TermList=ListOfPMOSTerminal
        Done=False
        while (TermList != None and not Done):
            if poly_touch(plist.Polygon,TermList.Polygon):
                if DevListActive.Device.term1 == None:
                    DevListActive.Device.term1=TermList.Polygon
                else:
                    if DevListActive.Device.term1 != TermList.Polygon:
                        DevListActive.Device.term3=TermList.Polygon
                        pntright=generate_pnt_right(DevListActive.Device.term2.head)
                        if point_inside_polygon(pntright, DevListActive.Device.term1) or point_inside_polygon(pntright,DevListActive.Device.term3):
                            DevListActive.Device.width=abs(DevListActive.Device.term2.head.pnt.x-DevListActive.Device.term2.head.next.pnt.x)
                            DevListActive.Device.width+=abs(DevListActive.Device.term2.head.pnt.y-DevListActive.Device.term2.head.next.pnt.y)
                            DevListActive.Device.length=abs(DevListActive.Device.term2.head.prev.pnt.x-DevListActive.Device.term2.head.pnt.x)
                            DevListActive.Device.length+=abs(DevListActive.Device.term2.head.prev.pnt.y-DevListActive.Device.term2.head.pnt.y)
                        else:
                            DevListActive.Device.length=abs(DevListActive.Device.term2.head.pnt.x-DevListActive.Device.term2.head.next.pnt.x)
                            DevListActive.Device.length+=abs(DevListActive.Device.term2.head.pnt.y-DevListActive.Device.term2.head.next.pnt.y)
                            DevListActive.Device.width=abs(DevListActive.Device.term2.head.prev.pnt.x-DevListActive.Device.term2.head.pnt.x)
                            DevListActive.Device.width+=abs(DevListActive.Device.term2.head.prev.pnt.y-DevListActive.Device.term2.head.pnt.y)                        
                        Done=True
            TermList=TermList.next
        plist=plist.next
        if plist is not None:
            DevListActive.next=DeviceList()
            DevListActive=DevListActive.next
            DevListActive.Device=Device()
            DevListActive.Device.name="PMOS"
        
    return DevList

def connected(name1,name2):
#
# DESCRIPTION: This routine checks if two layers are connected electrically
#
# INPUT: Two text fields: name1, name2;
#
# OUTPUT: True/False
#
# KNOWN ISSUES: The connectivity check is hardcoded. This should implemented
# as a matrix of sorts where rows and columns are identified with a particular
# layout layer
#
    if name1==name2:
        return True
    if name1=='Contact':
        if name2=='M0':
            return True
        else: 
            return False
    elif name1=="M0":
            if name2=="Contact" or name2=="VIA0":
                return True
            else: 
                return False
    elif name1=="VIA0":
        if name2=="M1" or name2=="M0":
            return True
        else:
            return False
    elif name1=="M1":
        if name2=="VIA0":
            return True
        else:
            return False
    elif name1=="Poly":
        if name2=="Contact" or name2=="NMOSbody" or name2=="PMOSbody":
            return True
        else:
            return False
    else:
        return False

def create_netlist(Cell):
#
# DESCRIPTION: This routine goes through all Devices in a class DeviceList and
# first initializes each device terminal with a unique class Net element. After
# this each "Contact" layer is looped trought to find those who touched the 
# device terminal, Term1. These "Contact" layers are now assigned the Term1 net.
# Then all polygons in "Poly" are looped thought to find those who touched the 
#Term2 terminal of the device. Lastly the "Contact" layer is looped through 
# again to find those touching Term3. Each time a "Contact" or "Poly" layer is
# found touching a device terminal the layer is assigned the terminal net.
#
# After this initial connectivity check all other layers are looped though one 
# by one to find those who touched "Contact" and "Poly" and so on. It assigns 
# the appropriate net if a polygon has no net assigned. If a polygon already 
# has a net attached to it the nets are joined
#
# INPUT: class Cell
#
# OUTPUT: None
#
# KNOWN ISSUES: No sanity checks. The short routine is not verified for many
# different situations. This implementation can easily be inadequate
#
    NetNumber=10000
    DevList=Cell.DeviceList
    while DevList is not None:
        Term1=DevList.Device.term1
        Term1.net=Net(NetNumber)
        NetNumber+=1
        ContactList=find_ListOfPolygonList(Cell, "Contact")
        while ContactList is not None:
            if poly_touch(ContactList.Polygon,Term1):
                ContactList.Polygon.net=Term1.net
            ContactList=ContactList.next
        Term2=DevList.Device.term2
        Term2.net=Net(NetNumber)
        NetNumber+=1
        ContactList=find_ListOfPolygonList(Cell, "Poly")
        while ContactList is not None:
            if poly_touch(ContactList.Polygon,Term2):
                if ContactList.Polygon.net is None:
                    ContactList.Polygon.net=Term2.net
                else:
                    if ContactList.Polygon.net.netNumber < Term2.net.netNumber:
                        Term2.net.next=ContactList.Polygon.net
                    else:
                        ContactList.Polygon.net.next=Term2.net
            ContactList=ContactList.next
        Term3=DevList.Device.term3
        Term3.net=Net(NetNumber)
        NetNumber=NetNumber+1
        ContactList=find_ListOfPolygonList(Cell, "Contact")
        while ContactList is not None:
            if poly_touch(ContactList.Polygon,Term3):
                ContactList.Polygon.net=Term3.net
            ContactList=ContactList.next                        
        DevList=DevList.next
    ListOfPolyList1=Cell.plist
    while ListOfPolyList1 is not None:
        ListOfPolyList2=Cell.plist
        while ListOfPolyList2 is not None:
            if connected(ListOfPolyList1.Name,ListOfPolyList2.Name):
                Plist1=ListOfPolyList1.PolyList
                while Plist1 is not None:
                    Plist2=ListOfPolyList2.PolyList
                    while Plist2 is not None:
                        if Plist1.Polygon is not Plist2.Polygon:
                            if poly_touch(Plist1.Polygon,Plist2.Polygon):
                                if Plist1.Polygon.net is not None:
                                    if Plist2.Polygon.net is not None:
                                        if Plist1.Polygon.net.netNumber < Plist2.Polygon.net.netNumber:
                                            Plist2.Polygon.net.next=Plist1.Polygon.net
                                        elif Plist2.Polygon.net.netNumber > Plist2.Polygon.net.netNumber:
                                            Plist1.Polygon.net.next=Plist2.Polygon.net
                                    else:
                                        Plist2.Polygon.net=Plist1.Polygon.net
                                else:
                                    if Plist2.Polygon.net is not None:
                                        Plist1.Polygon.net=Plist2.Polygon.net
                                    else:
                                        Plist1.Polygon.net=Net(NetNumber)
                                        Plist2.Polygon.net=Plist1.Polygon.net
                                        NetNumber=NetNumber+1
                        Plist2=Plist2.next
                    Plist1=Plist1.next
            ListOfPolyList2=ListOfPolyList2.next
        ListOfPolyList1=ListOfPolyList1.next

def print_netlist(Cell):
#
# DESCRIPTION: All devices found in class Cell are printed with the netnames
# on the terminals.
#
# INPUT: class Cell: Cell;
#
# OUTPUT: a printed netlist
#
# KNOWN ISSUES: If Cell is None it breaks.
#
    DevList=Cell.DeviceList
    while DevList is not None:
        term1Net=DevList.Device.term1.net
        while term1Net.next is not None:
            term1Net=term1Net.next
        term2Net=DevList.Device.term2.net
        while term2Net.next is not None:
            term2Net=term2Net.next
        term3Net=DevList.Device.term3.net
        while term3Net.next is not None:
            term3Net=term3Net.next
        print(term1Net.netNumber,' ',term2Net.netNumber,' ',term3Net.netNumber,' ',DevList.Device.name,' l=',DevList.Device.length,' w=',DevList.Device.width)
        DevList=DevList.next
    
    

