# -*- coding: utf-8 -*-
"""
Created on Wed Nov 17 12:09:12 2021

@author: msahr
"""
import matplotlib.pyplot as plt
import sys
import re
import math

class Nodes:
#
# DESCRIPTION: A Nodes class is an element that has two text field: Name,
# MactehdName, a boolean Matched and a connectivity class. It also has next
# and prev pointers for use in a double linked list.
#
    def __init__(self):

        self.Name=""
        self.MatchedName=""
        self.Matched=False
        self.Connectivity=[0 for i in range(6)]
        self.next=None
        self.prev=None
    def append(self,name):
#
# DESCRIPTION: This adds (appends) a node with Name=name to a class Nodes list
#
# INPUT: self, text: name;
# 
# OUTPUT: The class Nodes element containing the name: name
#
# KNOWN ISSUES: No sanity check
#
        NodesActive=self
        while NodesActive.next != None:
            if NodesActive.Name == name:
                return NodesActive
            NodesActive=NodesActive.next
        if NodesActive.Name == "":
            NodesActive.insertName(name)
        else:
            if NodesActive.Name == name:
                return NodesActive
            NodesActive.next=Nodes()
            NodesActive.next.prev=NodesActive
            NodesActive=NodesActive.next
            NodesActive.insertName(name)
        return NodesActive
    def insertName(self,name):
#
# DESCRIPTION: Changes self.Name to new: name
#
# INPUT: self, text: name;
#
# OUTPUT: NONE
#
# KNOWN ISSUES: There is no check if self is a Nodes object or if it has a Name
# assigned already
#
        self.Name=name
    def insertMatchedName(self,name):
#
# DESCRIPTION: Changes self.MatchedName to new: name
#
# INPUT: self, text: name;
#
# OUTPUT: NONE
#
# KNOWN ISSUES: There is no check if self is a Nodes object or if it has a 
# MatchedName assigned already
#        
        self.MatchedName=name
    def updateConnectivity(self,index):
#
# DESCRIPTION: updates the Connectivity array at index with one
# 
# INPUT: self, int: index;
#
# OUTPUT: None
#
# KNOWN ISSUES: No check if self is defined
#
        self.Connectivity[index]+=1
    def findNode(self,name):
#
# DESCRIPTION: Return the Nodes element with Name=name if found. If not found
# returns None
#
# INPUT: text: name;
#
# OUTPUT: class Nodes element 
#
# KNOWN ISSUES: None
#
        NodesActive=self
        while NodesActive is not None:
            if NodesActive.Name == name:
                return NodesActive
            NodesActive=NodesActive.next
        return None

def connectivityEqual(conn1,conn2):
#
# DESCRIPTION: If the two input connectivities are equal it return True, if not
# False
#
# INPUT:Connectivities: conn1, conn2;
#
# OUTPUT: True/False
#
# KNOWN ISSUES: No checks to see if conn1, conn2 are well defined
#
    for i in range(6):
        if conn1[i]!=conn2[i]:
            return False
    return True

   
class Devices:
#
# DESCRIPTION: The class Devices contains contains a text field Label that 
# represents the name of the device in a netlist. A modeIndex that indicates
# if it is a PMOS or NMOS device, three class Nodes pointers: Drain, Gate, Source;
# a text field MatchedLabel, a boolean variable: Matched indicating if a Device
# is matched to another and finally forward and backward pointers: next, prev;
    def __init__(self):
        self.Label=""
        self.modelIndex=0
        self.Drain=None
        self.Gate=None
        self.Source=None
        self.next=None
        self.prev=None
        self.MatchedLabel=""
        self.Matched=False
    def append(self,label,modelIndex,drain,gate,source):
#
# DESCRIPTION: This routine creates and appends a device with given properties
# to a class Device linked list.
#
# INPUT: self, text field: label; integer: modelIndex; class Nodes elements:
# drain, gate, source
#
# OUTPUT: The Device element with the input label, either created or found
#
# KNOWN ISSUES: Non sanity checks performed
#
        DevicesActive=self
        while DevicesActive.next != None:
            if DevicesActive.Label == label:
                return DevicesActive
            DevicesActive=DevicesActive.next
        if DevicesActive.Label != "":
            DevicesActive.next=Devices()
            DevicesActive.next.prev=DevicesActive
            DevicesActive=DevicesActive.next
        DevicesActive.Label=label
        DevicesActive.insertDrain(drain)
        DevicesActive.insertGate(gate)
        DevicesActive.insertSource(source)
        DevicesActive.modelIndex=modelIndex
        return DevicesActive
    def insertDrain(self,drain):
#
# DESCRIPTION: This assigns a pointer to self.drain
# 
# INPUT: self, class Nodes: drain;
#
# OUTPUT: None
#
# KNOWN ISSUES: No sanity check for the inputs
#
       self.Drain=drain
    def insertGate(self,gate):
#
# DESCRIPTION: This assigns a pointer to self.gate
# 
# INPUT: self, class Nodes: gate;
#
# OUTPUT: None
#
# KNOWN ISSUES: No sanity check for the inputs
#
       self.Gate=gate
    def insertSource(self,source):
#
# DESCRIPTION: This assigns a pointer to self.source
# 
# INPUT: self, class Nodes: source;
#
# OUTPUT: None
#
# KNOWN ISSUES: No sanity check for the inputs
#
       self.Source=source
       
def DeviceConnectivityEqual(Dev1,Dev2):
#
# DESCRIPTION: Verifies if two devices have the same terminal connectivities
#
# INPUT: class Device: Dev1, Dev2;
#
# OUTPUT: True/False
#
# KNOWN ISSUES: No sanity checks
#
    if connectivityEqual(Dev1.Drain.Connectivity, Dev2.Drain.Connectivity):
        if connectivityEqual(Dev1.Gate.Connectivity, Dev2.Gate.Connectivity):
            if connectivityEqual(Dev1.Source.Connectivity, Dev2.Source.Connectivity):
                return True
    return False
       
class Connectivity:
#
# DESCRIPTION: A Connectivity class consists of an array with six elements that
# annotates the number of times a NMOS&PMOS drain, gate, source are connected
# to a certain node. It also has an integer: NumberOfOccurences that shows how
# many times a certain connectivity occurs in the netlist.Finally pointers to
# the next and previous class Connectivity object are used.
#
    def __init__(self):
        self.Connectivity=[0 for i in range(6)]
        self.NumberOfOccurences=0
        self.next=None
        self.prev=None
    def append(self,Conn):
#
# DESCRIPTION: The append function looks through a Connectivity list and
# looks for a connectivity equal to the input: Conn; If one is found the 
# NumberOfOccurences field is updated. If not found a new Connectivity element
# is appended to the Connectivity list.
#
# INPUT: self, class Connectivity element: Conn;
#
# OUTPUT: None
#
# KNOWN ISSUES: The Connectivy field is hard coded to 6. In general this needs
# to be programmable
#
        SeekConn=[0 for i in range(6)]
        SeekConn[1]=1
        ConnActive=self
        if connectivityEqual(Conn, SeekConn):
            printIt=True
        else:
            printIt=False
        while ConnActive.next is not None:
            if connectivityEqual(ConnActive.Connectivity,Conn):
                ConnActive.NumberOfOccurences+=1
                return
            ConnActive=ConnActive.next
        if connectivityEqual(ConnActive.Connectivity,[0 for i in range(6)]):
            for i in range(6):
                ConnActive.Connectivity[i]=Conn[i]
            ConnActive.NumberOfOccurences=1
        else:
            if connectivityEqual(ConnActive.Connectivity,Conn):
                ConnActive.NumberOfOccurences+=1
            else:
                ConnActive.next=Connectivity()
                ConnActive.next.prev=ConnActive
                ConnActive=ConnActive.next
                for i in range(6):
                    ConnActive.Connectivity[i]=Conn[i]
                ConnActive.NumberOfOccurences=1
                
def characterizeConnectivities(Nodes,Connectivity):
#
# DESCRIPTION: A class Connectivity is updated with all nodes in a class Nodes
# linked list
#
# INPUT: a class Nodes: Nodes; a class Connectivity: Connectivity;
#
# OUTPUT: None
#
# KNOWN ISSUES: It assumes the nodes in Nodes has already had their Connectivty
# updated. No sanity checks
#
    NodesActive=Nodes
    while NodesActive is not None:
        Connectivity.append(NodesActive.Connectivity)
        NodesActive=NodesActive.next
        
def find_connectivity(Conn,Connection):
#
# DESCRIPTION: a Connectivity class: Conn; is looped through to find a certain
# connectivity given by input integer array: Connection. If found that Connectivity
# element is returned, if not it returns None
#
# INPUT: a class Connectivity: Conn, a 6 element integer array: Connection;
#
# OUTPUT: A pointer to a Connectivity class element 
#
# KNOWN ISSUES: No sanity checks
#
    connActive=Conn
    while connActive is not None:
        if connectivityEqual(connActive.Connectivity, Connection):
            return connActive
        connActive=connActive.next
    return None

def print_connectivity(Conn):
#
# DESCRIPTION: The input connectivity array is printed
#
# INPUT: A 6 element integer array
#
# OUTPUT: A printed connectivity
#
# KNOWN ISSUES: No sanity checks
#
    print('nmos drain:',Conn[0])
    print('nmos gate:',Conn[1])
    print('nmos source:',Conn[2])
    print('pmos drain:',Conn[3])
    print('pmos gate:',Conn[4])
    print('pmos source:',Conn[5])
            

def compareConnectivities(Conn1, Conn2):
#
# DESCRIPTION: The connectiviies of two class Connectivity elements are compared
# Discrepancies and similarities are printed
#
# INPUT: 2 class Connectivity elements
#
# OUTPUT: A human readable comparison of the two connectivities
#
# KNOWN ISSUES: No sanity checks
#
    connActive1=Conn1
    while connActive1 is not None:
        connActive2=find_connectivity(Conn2,connActive1.Connectivity)
        if connActive2 is not None:
            if connActive1.NumberOfOccurences!=connActive2.NumberOfOccurences:
                print('netlist 1 contains',connActive1.NumberOfOccurences,'node of Connectivity')
                print_connectivity(connActive1.Connectivity)
                print('while netlist 2 contains',connActive2.NumberOfOccurences,'such nodes')
        else:
            print('netlist 1 contains',connActive1.NumberOfOccurences,'node of Connectivity')
            print_connectivity(connActive1.Connectivity)
            print('while netlist 2 contains no such nodes')

        connActive1=connActive1.next

    connActive2=Conn2
    while connActive2 is not None:
        connActive1=find_connectivity(Conn1,connActive2.Connectivity)
        if connActive1 is not None:
            if connActive2.NumberOfOccurences!=connActive1.NumberOfOccurences:
                print('netlist 2 contains',connActive2.NumberOfOccurences,'node of Connectivity')
                print_connectivity(connActive2.Connectivity)
                print('while netlist 1 contains',connActive1.NumberOfOccurences,'such nodes')
        else:
            print('netlist 2 contains',connActive2.NumberOfOccurences,'node of Connectivity')
            print_connectivity(connActive2.Connectivity)
            print('while netlist 1 contains no such nodes')
        connActive2=connActive2.next
        
class MatchMatrixElement:
#
# DESCRIPTION: A MatchMatrixElement contains pointers to a class Nodes element
# four pointers to other MatchMatrixElement objects
#
    def __init__(self):
        self.Node=None
        self.down=None
        self.up=None
        self.left=None
        self.right=None
    def insertLeftNode(self,node):
#
# DESCRIPTION: A new left MatchMatrixElement is created (unless self.node==None)
# At the bottom of the Matrix given by self.
#
# INPUT: a MatchMatrixElement node: self; a class Nodes element:node;
#
# OUTPUT: A class MatchMatrixElement with the .node=input node
#
# KNOWN ISSUES: No sanity checks, for example self might not be the leftmost
# element (self.left != None!!), not checked.
#
        if self.Node==None:
            self.Node=node
            return self
        MatrixActive=self
        while MatrixActive.down is not None:
            MatrixActive=MatrixActive.down
        MatrixActive.down=MatchMatrixElement()
        MatrixActive.down.up=MatrixActive
        MatrixActive=MatrixActive.down
        MatrixActive.Node=node
        return MatrixActive
    def insertRightNode(self,node):
#
# DESCRIPTION: A new right MatchMatrixElement is created (unless self.node==None)
# At the rightmost end of 'self'.
#
# INPUT: a MatchMatrixElement node: self; a class Nodes element:node;
#
# OUTPUT: A class MatchMatrixElement with the .node=input node
#
# KNOWN ISSUES: No sanity checks.
#
        # Here it is assumed self is an already defined left-node
        if self.Node==None:
            print('Error: THe call to insertRightNode should have an already defined left node in its argument')
            sys.exit()
        MatrixActive=self
        while MatrixActive.right is not None:
            MatrixActive=MatrixActive.right
        MatrixActive.right=MatchMatrixElement()
        MatrixActive.right.left=MatrixActive
        MatrixActive=MatrixActive.right
        MatrixActive.Node=node
        return
    def insertLeftDevice(self,Device):
#
# DESCRIPTION: A full device with all three terminals is inserted in one shot
# on the left MatchMatrix Column
#
# INPUT: a MatchMatrixElement: self; a class Devices: Device; 
#
# OUTPUT: A class MatchMatrixElement pointing to the node given by Device.drain
#
# KNOWN ISSUES: More sanity checks needed
#
        if self.Node is not None:
            print('insertLeftDevice error: The assumption is that self contains the new matrix element where the first device port is to be linked')
            sys.exit()
        self.Node=Device.Drain
        self.down=MatchMatrixElement()
        self.down.up=self
        self.down.Node=Device.Gate
        self.down.down=MatchMatrixElement()
        self.down.down.up=self.down
        self.down.down.Node=Device.Source
        return self
    def insertRightDevice(self,Device):
#
# DESCRIPTION: A full device with all three terminals is inserted in one shot
# along the right linked list of a MatchMatrix Row
#
# INPUT: a MatchMatrixElement: self; a class Devices: Device; 
#
# OUTPUT: A class MatchMatrixElement pointing to the node given by Device.drain
#
# KNOWN ISSUES: We assume the self pointer points to the left side upper 
# port of a device
#
        if self.Node == None or self.down == None or self.down.down == None:
            print('Error: THe call to insertRightDevice should have an already defined device node in its argument')
            sys.exit()
        port1=self
        port2=self.down
        port3=self.down.down
        while port1.right is not None:
            port1=port1.right
            port2=port2.right
            port3=port3.right
            # Error checks needed!
        port1.right=MatchMatrixElement()
        port2.right=MatchMatrixElement()
        port3.right=MatchMatrixElement()
        
        port1.right.left=port1
        port2.right.left=port2
        port3.right.left=port3
        
        port1=port1.right
        port2=port2.right
        port3=port3.right
        
        port1.down=port2
        port2.up=port1
        port2.down=port3
        port3.up=port2
        
        port1.Node=Device.Drain
        port2.Node=Device.Gate
        port3.Node=Device.Source
    def deleteDevice(self,port):
#
# DESCRIPTION: A device section of a MatchMatrix is deleted from the rhs side
#
# INPUT: self, class MatchMatrixElement: port;
#
# OUTPUT: None
#
# KNOWN ISSUES: No sanity checks
#
        portActive=port
        while portActive.up is not None:
            portActive=portActive.up
        while portActive is not None:
            portActive.left.right=portActive.right
            if portActive.right is not None:
                portActive.right.left=portActive.left
            portActive=portActive.down
        return
    def deleteNode(self,node):
#
# DESCRIPTION: A MatchMatrix is deleted from the rhs side
#
# INPUT: self, class MatchMatrixElement: node;
#
# OUTPUT: None
#
# KNOWN ISSUES: No sanity checks
#
        node.left.right=node.right
        if node.right is not None:
            node.right.left=node.left
    def print(self):
#
# DESCRIPTION: a MatchMatrixElement given by self is printed
# 
# INPUT: self
#
# OUTPUT: A MatchMatrix in human readable form
#
# KNOWN ISSUES: No sanity checks
#
        MatrixLeft=self
        while MatrixLeft is not None:
            print(MatrixLeft.Node.Name,'->',end='')
            MatrixRight=MatrixLeft.right
            while MatrixRight is not None:
                print('|',MatrixRight.Node.Name,sep='',end='')
                MatrixRight=MatrixRight.right
            print('>')
            MatrixLeft=MatrixLeft.down
            
def buildMatchMatrix(MatchMatrix,Devices1,Nodes1,Devices2,Nodes2):
#
# DESCRIPTION: Given a MatchMatrix element and device and nodes linked lists
# from two netlists a new MatchMatrix is built-up
#
# INPUT: a MatchMatrix element: MatchMatrix; class Devices: Devices1,
# Devices2; class Nodes: Nodes1, Nodes2;
#
# OUTPUT: Nonde
#
# KNOWN ISSUES: No sanity checks
#
    Nodes1Active=Nodes1
    while Nodes1Active is not None:
        LeftElement=MatchMatrix.insertLeftNode(Nodes1Active)
        Nodes2Active=Nodes2
        while Nodes2Active is not None:
            if connectivityEqual(Nodes1Active.Connectivity, Nodes2Active.Connectivity):
                LeftElement.insertRightNode(Nodes2Active)
            Nodes2Active=Nodes2Active.next
        Nodes1Active=Nodes1Active.next
    MatrixActive=MatchMatrix
    while MatrixActive.down is not None:
        MatrixActive=MatrixActive.down
    MatrixActive.down=MatchMatrixElement()
    MatrixActive.down.up=MatrixActive
    MatrixActive=MatrixActive.down
    Devices1Active=Devices1
    while Devices1Active is not None:
        MatrixActive=MatrixActive.insertLeftDevice(Devices1Active)
        Devices2Active=Devices2
        while Devices2Active is not None:
            if DeviceConnectivityEqual(Devices1Active,Devices2Active):
                MatrixActive.insertRightDevice(Devices2Active)
            Devices2Active=Devices2Active.next
        Devices1Active=Devices1Active.next
        if Devices1Active is not None:
            while MatrixActive.down is not None:
                MatrixActive=MatrixActive.down
            MatrixActive.down=MatchMatrixElement()
            MatrixActive.down.up=MatrixActive
            MatrixActive=MatrixActive.down
              
def MatrixAnd(Matrix):
#
# DESCRIPTION: A matrix AND operation is performed by looping through the left
# column of a Matrix and comparing the match for all same nodes along the left
# column. Any match on the right side of Matrix that does not occur for all same
# nodes is removed
#
# INPUT: a Connectivity Matrix: Matrix;
#
# OUTPUT: None
#
# KNOWN ISSUES: No Sanity checks
#
    StillChanging=True
    while StillChanging:
        StillChanging=False
        MatrixLeft=Matrix
        while MatrixLeft is not None:
            MatrixLeft2=Matrix
            while MatrixLeft2 is not None:
                if MatrixLeft2 != MatrixLeft:
                    if MatrixLeft.Node.Name == MatrixLeft2.Node.Name:
                        MatrixRight=MatrixLeft.right
                        while MatrixRight is not None:
                            MatrixRight2=MatrixLeft2.right
                            FoundNameIn2=False
                            while MatrixRight2 is not None:
                                if MatrixRight.Node.Name == MatrixRight2.Node.Name:
                                    FoundNameIn2=True
                                MatrixRight2=MatrixRight2.right
                            if not FoundNameIn2:
                                StillChanging=True
                                MatrixRight.deleteDevice(MatrixRight)
                            MatrixRight=MatrixRight.right
                MatrixLeft2=MatrixLeft2.down
            MatrixLeft=MatrixLeft.down
            
def RemoveMatchedRightNodesMatrix(Matrix):
#
# DESCRIPTION: A MatchMatrix is looped through and if a Matched node is found
# on the left column all right Matrix elements are compared to the MatchedName
# If the MatchedName is not found, the full device is removed
#
# INPUT: a class MatchMatrixElement: Matrix;
#
# OUTPUT: None
#
# KNOWN ISSUES: No Sanity checks
#
    MatrixActive=Matrix
    Print=False
    while MatrixActive is not None:
        MatrixRight=MatrixActive.right
        while MatrixRight is not None:
            # Debug code snippet:
            # if MatrixActive.Node.Name=='Iout2':
            #     Print=False
            # else:
            #     Print=False
            if Print:
                print('Node is matched',MatrixRight.Node.Matched,MatrixRight.Node.Name)
            if MatrixRight.Node.Matched:
                if Print:
                    print("Right",MatrixActive.Node.MatchedName,MatrixRight.Node.Name)
                if MatrixActive.Node.MatchedName != MatrixRight.Node.Name:
                    MatrixRight.deleteDevice(MatrixRight)
            MatrixRight=MatrixRight.right
        MatrixActive=MatrixActive.down
        
            
def MatrixMatch(Matrix):
# 
# DESCRIPTION: A matrix is looped through looking for matches. If a left column
# has right elements with the same unique nodes, the left element node is considered
# matched to the right node
#
# INPUT: a class MatchMatrixElement: Matrix;
#
# OUTPUT: None
#
# KNOWN ISSUES: No sanity checks
#
    MatrixActive=Matrix
    while MatrixActive is not None:
        MatrixRight=MatrixActive.right
        if MatrixActive.Node.Matched:
            while MatrixRight is not None:
                if MatrixActive.Node.MatchedName != MatrixRight.Node.Name:
                    MatrixRight.deleteDevice(MatrixRight)
                MatrixRight=MatrixRight.right
        elif MatrixRight is not None:
            if MatrixRight.right==None:
                MatchNodes(MatrixActive.Node,MatrixRight.Node)
        MatrixActive=MatrixActive.down
       
def MatchNodes(node1, node2):
#
# DESCRIPTION: Two nodes are matched by causing their MatchedName to point to
# the others Name and the Nodes.Matched=True for both nodes
#
# INPUT: two class Nodes: node1, node2;
#
# OUTPUT: None
#
# KNOWN ISSUES: No sanity checks
#
    print('Macthing',node1.Name,'to',node2.Name)
    node1.Matched=True
    node2.Matched=True
    node1.MatchedName=node2.Name
    node2.MatchedName=node1.Name
    
def SymmetryMatch(Matrix):
#
# DESCRIPTION: A symmetry match is performed by taking the first matrix row
# that has at least two possible matched on the right, and picking one of these
# as a match. If a match is found it returns True, otherwise False
#
# INPUT: a class MatchMatrixElement: Matrix;
#
# OUTPUT: A boolean True/False
#
# KNOWN ISSUES: No sanity checks
#
    MatrixActive=Matrix
    FoundSymmetryPoint=False
    while not FoundSymmetryPoint and MatrixActive is not None:
        # print('MatrixActive',MatrixActive.Node.Name)
        if MatrixActive.right is not None:
            if MatrixActive.right.right is not None:
                FoundSymmetryPoint=True
                print('SymmetryMatch!')
                MatchNodes(MatrixActive.Node,MatrixActive.right.Node)
                MatrixRight=MatrixActive.right.right
                while MatrixRight is not None:
                    MatrixRight.deleteDevice(MatrixRight)
                    MatrixRight=MatrixRight.right
        MatrixActive=MatrixActive.down
    if not FoundSymmetryPoint:
        print('No Symmetry Point found in Matrix')
    return FoundSymmetryPoint

def DeviceMatch(Devs1,Devs2):
#
# DESCIPTION: Two class Devices lists are compared. If the terminals of two devices
# are matched the devices are considered matched and Device.Matched=True and 
# Device.MatchedLabel=Label If matched devices are found it returns True if not
# False
#
# INPUT: two classes Devices: Devs1, Devs2;
#
# OUTPUT: True/False
#
# KNOWN ISSUES: No sanity checks
#
    Devs1Active=Devs1
    NetlistsMatch=True
    while Devs1Active is not None:
        Devs2Active=Devs2
        FoundMatchedDevice=False
        while not FoundMatchedDevice and Devs2Active is not None:
            if Devs1Active.Drain.MatchedName == Devs2Active.Drain.Name:
                if Devs1Active.Gate.MatchedName == Devs2Active.Gate.Name:
                    if Devs1Active.Source.MatchedName == Devs2Active.Source.Name:
                        if not Devs2Active.Matched:
                            FoundMatchedDevice=True
                            print('Matching',Devs1Active.Label,'to',Devs2Active.Label)
                            Devs1Active.MatchedLabel=Devs2Active.Label
                            Devs1Active.Matched=True
                            Devs2Active.Matched=True
            Devs2Active=Devs2Active.next
        if not FoundMatchedDevice:
            print('No match found for device',Devs1Active.Label)
            NetlistsMatch=False
        Devs1Active=Devs1Active.next
    return NetlistsMatch
    
def clearDeviceMatch(Devs):
#
# DESCRIPTION: All Devices.Matched are set to False in a class Devices list
#
# INPUT: class Devices: Devs;
#
# OUTPUT: None
#
# KNOWN ISSUES: No sanity checks
#
    DevsActive=Devs
    while DevsActive is not None:
        DevsActive.Matched=False
        DevsActive=DevsActive.next
        
def characterize_netlist(Devs,Nets):
#
# DESCRIPTION: The number of NMOS/PMOS found in devices and the number of nets
# in Nets is counted
#
# INPUT: class Devices: Devs; class Nodes: Nets;
#
# OUTPUT: The number of NMOS/PMOS/Nodes
#
# KNOWN ISSUES:No sanity checks
#
    Npmos=Nnmos=Nnodes=0
    DevsActive=Devs
    while DevsActive is not None:
        if DevsActive.modelIndex==0:
            Nnmos+=1
        else:
            Npmos+=1
        DevsActive=DevsActive.next
    NetsActive=Nets
    while NetsActive is not None:
        Nnodes+=1
        NetsActive=NetsActive.next
    return Nnmos,Npmos,Nnodes
        
def readnetlist(netlist,Nodes, Devices):
#
# DESCRIPTION: A netlist given by in input file is read and the found Nodes
# and devices are returned as linked lists.
#
# INPUT: a text: netlist; class Nodes: Nodes; class Devices: Devices;
#
# OUTPUT: The number of devices found
#
# KNOWN ISSUES: The syntax for the input format cannot deviate much from
# a CDL type netlist. Things like spaces need to be happening only one at a time
#
    try:
        myfile=open(netlist,'r')
    except:
        print('netlist file',netlist,' not found')
        sys.exit()
    DeviceCount=0
    line=myfile.readline()
    while line !='' :
        DevType='empty'
        if line[0]=='*': 
            print('comment')
        if line[0]=='m' or line[0]=='M':
            DevType='transistor'
        if line[0]==' ':
            print('Empty beginning of line not allowed')
            sys.exit()
        if DevType!='empty':
                try:
                    DevLabel=line.split(' ')[0]
                except: 
                    print('Syntax Error in line:',line) 
                    sys.exit();
                try:
                    Drain=line.split(' ')[1]
                except: 
                    print('Syntax Error in line:',line) 
                    sys.exit();
                try:
                    Gate=line.split(' ')[2]
                except: 
                    print('Syntax Error in line:',line) 
                    sys.exit();
                try:
                    Source=line.split(' ')[3]
                except: 
                    print('Syntax Error in line:',line) 
                    sys.exit();
                try:
                    DevModel=line.split(' ')[4]
                    DevModel=DevModel.rstrip('\n')
                    if DevModel == 'nmos':
                        print('Found NMOS device')
                        modelIndex=0
                    elif DevModel == 'pmos':
                        modelIndex=1
                        print('Found PMOS device')
                    else:
                        print('Error, only nmos or pmos models allowed')
                        sys.exit()
                except: 
                    print('Syntax Error in line4:',line) 
                    sys.exit();
                NodesActive=Nodes.append(Drain)
                NodesActive.updateConnectivity(modelIndex*3+0)
                NodesActive=Nodes.append(Gate)
                NodesActive.updateConnectivity(modelIndex*3+1)
                NodesActive=Nodes.append(Source)
                NodesActive.updateConnectivity(modelIndex*3+2)
                Devices.append(DevLabel,modelIndex,Nodes.findNode(Drain),Nodes.findNode(Gate),Nodes.findNode(Source))
                DeviceCount+=1 
        line=myfile.readline()
    return DeviceCount

def CompareNetlists_barebones(netlist1, netlist2):
#
# DESCRIPTION: A very basic netlist comparison is made with no debugging output
#
# INPUT: Two netlists file names: netlist1, netlist2;
#
# OUTPUT: A printed comparison and a concluding Matched/not Matched result
#
# KNOWN ISSUES: None
#
    Devs1=Devices()
    Nets1=Nodes()
    Devs2=Devices()
    Nets2=Nodes()
    Matrix=MatchMatrixElement()
    
    print('reading netlist 1')
    # readnetlist('netlist8p3a.txt',Nets1,Devs1)
    readnetlist(netlist1,Nets1,Devs1)
    print('...done')
    print('readin netlist 2')
    # readnetlist('netlist8p3b.txt',Nets2,Devs2)
    readnetlist(netlist2,Nets2,Devs2)
    print('... done')
    buildMatchMatrix(Matrix,Devs1,Nets1,Devs2,Nets2)
    # Matrix.print()
    print(' ')
    print('Macthing ...')
    MatrixMatch(Matrix)
    # Matrix.print()
    print('... done')
    print('MatrixAND ...')
    MatrixAnd(Matrix)
    # Matrix.print()
    print('... done')
    print('Macthing ...')
    MatrixMatch(Matrix)
    FindingSymmetryPoints=True
    # Matrix.print()
    while FindingSymmetryPoints:
        FindingSymmetryPoints=SymmetryMatch(Matrix)
        print('Clean up Matrix')
        RemoveMatchedRightNodesMatrix(Matrix)
        #Matrix.print()
        MatrixAnd(Matrix)
        MatrixMatch(Matrix)
        #Matrix.print()
        print('Clean up Matrix')
        RemoveMatchedRightNodesMatrix(Matrix)
        #Matrix.print()
        MatrixMatch(Matrix)
        print('Clean up Matrix')
        RemoveMatchedRightNodesMatrix(Matrix)
        # Matrix.print()
    print('... done')
    # Matrix.print()
    print('Verifying Isomorphism, netlist1->netlist2 ...')
    MatchFirst=DeviceMatch(Devs1,Devs2)
    clearDeviceMatch(Devs1)
    clearDeviceMatch(Devs2)
    if not MatchFirst:
        print('Netlists do NOT match')
    print('... and netlist2->netlist1')
    MatchSecond=DeviceMatch(Devs2,Devs1)
    if not MatchSecond:
        print('Netlists do NOT match')
    if MatchFirst and MatchSecond:
        print('Netlists Match')
    print('done')

def CompareNetlists_simple(netlist1, netlist2):
#
# DESCRIPTION: A very basic netlist comparison is made with some debugging output
#
# INPUT: Two netlists file names: netlist1, netlist2;
#
# OUTPUT: A printed comparison and a concluding Matched/not Matched result
#
# KNOWN ISSUES: None
#
    Devs1=Devices()
    Nets1=Nodes()
    Devs2=Devices()
    Nets2=Nodes()
    Matrix=MatchMatrixElement()
    
    print('reading netlist 1')
    # readnetlist('netlist8p3a.txt',Nets1,Devs1)
    readnetlist(netlist1,Nets1,Devs1)
    print('...done')
    print('readin netlist 2')
    # readnetlist('netlist8p3b.txt',Nets2,Devs2)
    readnetlist(netlist2,Nets2,Devs2)
    print('... done')
    Nnmos1,Npmos1,Nnodes1=characterize_netlist(Devs1,Nets1)
    Nnmos2,Npmos2,Nnodes2=characterize_netlist(Devs2,Nets2)
    if Nnmos1 == Nnmos2:
        print('The netlists contain',Nnmos1,'nmos devices each')
    else:
        print('The netlists contain different numbers of nmos devices')
        print('\tnetlist 1 contains',Nnmos1,'while netlist 2 contains',Nnmos2,'nmos devices')
    if Npmos1 == Npmos2:
        print('The netlists contain',Npmos1,'pmos devices each')
    else:
        print('The netlists contain different numbers of pmos devices:')
        print('\tnetlist 1 contains',Npmos1,'while netlist 2 contains',Npmos2,'pmos devices')
    if Nnodes1 == Nnodes2:
        print('The netlists contain',Nnodes1,'nodes each')
    else:
        print('The netlists contain different number of nodes')
        print('\tnetlist 1 contains',Nnodes1,'while netlist 2 contains',Nnodes2,'nodes')

    buildMatchMatrix(Matrix,Devs1,Nets1,Devs2,Nets2)
    # Matrix.print()
    print(' ')
    print('Macthing ...')
    MatrixMatch(Matrix)
    # Matrix.print()
    print('... done')
    print('MatrixAND ...')
    MatrixAnd(Matrix)
    # Matrix.print()
    print('... done')
    print('Macthing ...')
    MatrixMatch(Matrix)
    FindingSymmetryPoints=True
    # Matrix.print()
    while FindingSymmetryPoints:
        FindingSymmetryPoints=SymmetryMatch(Matrix)
        print('Clean up Matrix')
        RemoveMatchedRightNodesMatrix(Matrix)
        #Matrix.print()
        MatrixAnd(Matrix)
        MatrixMatch(Matrix)
        #Matrix.print()
        print('Clean up Matrix')
        RemoveMatchedRightNodesMatrix(Matrix)
        #Matrix.print()
        MatrixMatch(Matrix)
        print('Clean up Matrix')
        RemoveMatchedRightNodesMatrix(Matrix)
        # Matrix.print()
    print('... done')
    # Matrix.print()
    print('Verifying Isomorphism, netlist1->netlist2 ...')
    MatchFirst=DeviceMatch(Devs1,Devs2)
    clearDeviceMatch(Devs1)
    clearDeviceMatch(Devs2)
    if not MatchFirst:
        print('Netlists do NOT match')
    print('... and netlist2->netlist1')
    MatchSecond=DeviceMatch(Devs2,Devs1)
    if not MatchSecond:
        print('Netlists do NOT match')
    if MatchFirst and MatchSecond:
        print('Netlists Match')
    print('done')
    
def CompareNetlists(netlist1, netlist2):
#
# DESCRIPTION: A very basic netlist comparison is made with more debugging output
#
# INPUT: Two netlists file names: netlist1, netlist2;
#
# OUTPUT: A printed comparison and a concluding Matched/not Matched result
#
# KNOWN ISSUES: None
#
    Devs1=Devices()
    Nets1=Nodes()
    Devs2=Devices()
    Nets2=Nodes()
    Matrix=MatchMatrixElement()
    
    print('reading netlist 1')
    # readnetlist('netlist8p3a.txt',Nets1,Devs1)
    readnetlist(netlist1,Nets1,Devs1)
    print('...done')
    print('readin netlist 2')
    # readnetlist('netlist8p3b.txt',Nets2,Devs2)
    readnetlist(netlist2,Nets2,Devs2)
    print('... done')
    Nnmos1,Npmos1,Nnodes1=characterize_netlist(Devs1,Nets1)
    Nnmos2,Npmos2,Nnodes2=characterize_netlist(Devs2,Nets2)
    if Nnmos1 == Nnmos2:
        print('The netlists contain',Nnmos1,'nmos devices each')
    else:
        print('The netlists contain different numbers of nmos devices')
        print('\tnetlist 1 contains',Nnmos1,'while netlist 2 contains',Nnmos2,'nmos devices')
    if Npmos1 == Npmos2:
        print('The netlists contain',Npmos1,'pmos devices each')
    else:
        print('The netlists contain different numbers of pmos devices:')
        print('\tnetlist 1 contains',Npmos1,'while netlist 2 contains',Npmos2,'pmos devices')
    if Nnodes1 == Nnodes2:
        print('The netlists contain',Nnodes1,'nodes each')
    else:
        print('The netlists contain different number of nodes')
        print('\tnetlist 1 contains',Nnodes1,'while netlist 2 contains',Nnodes2,'nodes')
    
    print('Characterize connectivities')
    Connectivity1=Connectivity()
    characterizeConnectivities(Nets1, Connectivity1)
    Connectivity2=Connectivity()
    characterizeConnectivities(Nets2, Connectivity2)
    
    compareConnectivities(Connectivity1, Connectivity2)


    buildMatchMatrix(Matrix,Devs1,Nets1,Devs2,Nets2)
    # Matrix.print()
    print(' ')
    print('Macthing ...')
    MatrixMatch(Matrix)
    # Matrix.print()
    print('... done')
    print('MatrixAND ...')
    MatrixAnd(Matrix)
    # Matrix.print()
    print('... done')
    print('Macthing ...')
    MatrixMatch(Matrix)
    FindingSymmetryPoints=True
    # Matrix.print()
    while FindingSymmetryPoints:
        FindingSymmetryPoints=SymmetryMatch(Matrix)
        print('Clean up Matrix')
        RemoveMatchedRightNodesMatrix(Matrix)
        #Matrix.print()
        MatrixAnd(Matrix)
        MatrixMatch(Matrix)
        #Matrix.print()
        print('Clean up Matrix')
        RemoveMatchedRightNodesMatrix(Matrix)
        #Matrix.print()
        MatrixMatch(Matrix)
        print('Clean up Matrix')
        RemoveMatchedRightNodesMatrix(Matrix)
        # Matrix.print()
    print('... done')
    # Matrix.print()
    print('Verifying Isomorphism, netlist1->netlist2 ...')
    MatchFirst=DeviceMatch(Devs1,Devs2)
    clearDeviceMatch(Devs1)
    clearDeviceMatch(Devs2)
    if not MatchFirst:
        print('Netlists do NOT match')
    print('... and netlist2->netlist1')
    MatchSecond=DeviceMatch(Devs2,Devs1)
    if not MatchSecond:
        print('Netlists do NOT match')
    if MatchFirst and MatchSecond:
        print('Netlists Match')
    print('done')

